/* This program is copyrighted, but may be used in terms of the GPL  */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <values.h>
#include <ctype.h>
#include <signal.h>

#include <linux/kd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#include <kunf.h>

/*********************************************************************/

#define LOADAVG  "/proc/loadavg"
#define LBUF    1024

#define SOCKSTAT "/proc/net/sockstat"
#define SBUF    1024

#define MEMINFO  "/proc/meminfo"
#define MBUF    1024

#define CONSOLE  "/dev/console"

#define LOCKFILE "/var/run/ecgd.pid"

/* Fiddle the values below *******************************************/

#define CRITICAL  2.5     /* cpu load above which ecgd goes to sleep */
#define PAUSETIME   5     /* time for which ecgd sleeps              */

/* misleading - M??FREQ are not frequencies at all *******************/

#define MAXFREQ   600     /* sound at highest load (when flat-lining)*/
#define MINFREQ 25000     /* sound at lowest lowest load             */

#define SLEEPTIME 400000  /* default time between beeps (in micro s) */

/* Do not fiddle the values below here *******************************/

int useleds;          /* blinkende lichter                           */
int usespeaker;       /* beeping speaker                             */
int force;            /* force loading                               */

/* lots of ideas, but unimplemented... we could watch syslog, mail.. */

int doload;
int dosocket;
int docore;

int domail; 
int dologin;

unsigned long sleeptime;  /* time between updates                        */
int lowerprio;            /* nice level                                  */
int sleeponlogin;         /* stop messing around if someone uses console */
int trapsig;              /* beep if received USR1 or USR2 signal ?      */
int readfifo;             /* read instructions from fifo                 */
int flatline;             /* do nothing if system is suffering           */

double arrestload;        /* quit if worse than this                     */

FILE *cfp;

#define PRIMARY   "monitor"
#define SECONDARY "ecgd"

#define BEEPING   "sound"
#define LEDS      "light"
#define CRITICALL "critical_load"
#define FORCE     "force_loading"
#define NICE      "nice_level"
#define PERIOD    "cycle_period"

int readkunfig(){
  char *str;

  kunfig_open(NULL,KUNFIG_OPEN_DEFAULTINGENV|KUNFIG_OPEN_SIMPLEACCESS|KUNFIG_OPEN_EXITONERRORS|KUNFIG_OPEN_LOGERRORS);

  str=kunfig_findvalue(3,PRIMARY,SECONDARY,BEEPING);
  if(str!=NULL){
    str[0]=tolower(str[0]);
    if(str[0]=='y'||str[0]=='e'){
      usespeaker=1;
    }
    else{
      usespeaker=0;
    }
  }
  str=kunfig_findvalue(3,PRIMARY,SECONDARY,LEDS);
  if(str!=NULL){
    str[0]=tolower(str[0]);
    if(str[0]=='y'||str[0]=='e'){
      useleds=1;
    }
    else{
      useleds=0;
    }
  }
  str=kunfig_findvalue(3,PRIMARY,SECONDARY,FORCE);
  if(str!=NULL){
    str[0]=tolower(str[0]);
    if(str[0]=='y'||str[0]=='e'){
      force=1;
    }
    else{
      force=0;
    }
  }

  str=kunfig_findvalue(3,PRIMARY,SECONDARY,CRITICALL);
  if(str!=NULL){
    arrestload=atof(str);
  }
  str=kunfig_findvalue(3,PRIMARY,SECONDARY,NICE);
  if(str!=NULL){
    lowerprio=atoi(str);
  }
  str=kunfig_findvalue(3,PRIMARY,SECONDARY,PERIOD);
  if(str!=NULL){
    sleeptime=atol(str);
  }

  kunfig_close();
  return 0;
}

int meminfo(int *a, int *b, int *c, int *d){
  FILE *pfp;
  char  pbuf[MBUF];
  int   i,ram=0,used=0,swap=0,paged=0;

  pfp=fopen(MEMINFO,"r");
  if(pfp!=NULL){
    fgets(pbuf,LBUF,pfp);
    fgets(pbuf,LBUF,pfp);
    i=0;
    while(!isdigit(pbuf[i])&&(pbuf[i]!='\0')){
      i++;
    }
    ram=atoi(pbuf+i);
    while(isdigit(pbuf[i])&&(pbuf[i]!='\0')){
      i++;
    }
    i++;
    used=atoi(pbuf+i);

#ifdef TRACE
fprintf(stderr,"Using <%d> of <%d> bytes\n",used,ram);
#endif

    fgets(pbuf,LBUF,pfp);
    i=0;
    while(!isdigit(pbuf[i])&&(pbuf[i]!='\0')){
      i++;
    }
    swap=atoi(pbuf+i);
    while(isdigit(pbuf[i])&&(pbuf[i]!='\0')){
      i++;
    }
    i++;
    paged=atoi(pbuf+i);

#ifdef TRACE
fprintf(stderr,"Using <%d> of <%d> bytes of swap\n",paged,swap);
#endif

    fclose(pfp);
  }

  *a=ram;
  *b=used;
  *c=swap;
  *d=paged;
   
  return 0;
}

int sockstat(int *a){
  FILE *pfp;
  char  pbuf[SBUF];
  int   i,count=0;

  pfp=fopen(SOCKSTAT,"r");
  if(pfp!=NULL){
    fgets(pbuf,LBUF,pfp);
    fclose(pfp);
    i=0;
    while(!isdigit(pbuf[i])&&(pbuf[i]!='\0')){
      i++;
    }
    count=atoi(pbuf+i);
#ifdef TRACE
fprintf(stderr,"Using <%d> sockets\n",count);
#endif
  }
  
  *a=count; 
  return 0;
}

int loadavg(double *a, double *b, double *c){
  FILE *pfp;
  char  pbuf[LBUF];
  char *ptr;

  pfp=fopen(LOADAVG,"r");
  if(pfp!=NULL){
    fgets(pbuf,LBUF,pfp);

    ptr=pbuf;

    *a=atof(ptr);
    while(!isspace(ptr[0]))ptr++;
    while( isspace(ptr[0]))ptr++;
    *b=atof(ptr);
    while(!isspace(ptr[0]))ptr++;
    while( isspace(ptr[0]))ptr++;
    *c=atof(ptr);
      
#ifdef TRACE
fprintf(stderr,"Load is <%f><%f><%f>\n",*a,*b,*c);
#endif

    fclose(pfp);
  } 

  return 0;
}

static void trapkill(int sig){

  if(usespeaker){
    ioctl(fileno(cfp),KIOCSOUND,0);
  }
  if(useleds){
    ioctl(fileno(cfp),KDSETLED,0);
  }

  fclose(cfp);

  unlink(LOCKFILE);

  exit(0);
}





int main(int argc, char **argv){

  double maxload,minload;
  double load,prev,past; 

  int    maxsock,minsock;
  int    sock,oldsock;
 
  int    maxused,minused,maxpaged,minpaged;
  int    used,paged;
 
  int    ram,swap;

  int    consolenum; 
  int    ioval;


  struct stat fs;

  char *ptr;
  int  i,j,m=0;
  double f;
  int niceval=1;

  sleeptime =SLEEPTIME;
  useleds   =1;
  usespeaker=1;
  domail    =1; 
  doload    =1;
  dologin   =1;
  dosocket  =1;
  lowerprio =0;

  arrestload=CRITICAL;
  flatline  =PAUSETIME;

  maxload=(-0.99*MAXDOUBLE);
  minload=MAXDOUBLE;

  maxsock=0;
  minsock=MAXINT;
  oldsock=MAXINT;

  maxused=0;
  minused=MAXINT;

  maxpaged=0;
  minpaged=0;

  force=0;
  ioval=0;

  readkunfig();

/* parse command-line */
  i=1;
  while(i<argc){
    ptr=argv[i];
    while(ptr[0]!='\0'){
      switch(ptr[0]){
        case '-' : 
        break;
        case 'L' : useleds=0;
        break;
        case 'B' : usespeaker=0;
        break;
        case 'f' : force=1;
        break;
        case 'm' : m=1;
        break;
        case 'n' : ptr++;
          if(ptr[0]=='\0'){
            i++;
            if(i<argc)ptr=argv[i];
          }
          niceval=atoi(ptr);
          while(ptr[1]!='\0'&&ptr[0]!='\0')ptr++;
        break;
        case 'p' : ptr++;
          if(ptr[0]=='\0'){
            i++;
            if(i<argc)ptr=argv[i];
          }
          sleeptime=atoi(ptr);
          while(ptr[1]!='\0'&&ptr[0]!='\0')ptr++;
        break;
        case 'c' : ptr++;
          if(ptr[0]=='\0'){
            i++;
            if(i<argc)ptr=argv[i];
          }
          arrestload=atof(ptr);
          while(ptr[1]!='\0'&&ptr[0]!='\0')ptr++;
        break;
        case 'h' : 
fprintf(stderr,"\n    ECG for Linux\n");
fprintf(stderr,"Usage: %s [-B] [-L] [-c <float>] [-f] [-n <int>]\n\n",argv[0]);
fprintf(stderr,"-B         Disable Beeping (only flash lights)\n");
fprintf(stderr,"-L         Disable Lights  (only do beeping)\n");
fprintf(stderr,"-c <float> Load above which %s goes to sleep. Defaults to <%f>\n",argv[0],CRITICAL);
fprintf(stderr,"-f         Force loading (even if <%s> exists)\n",LOCKFILE);
fprintf(stderr,"-n <int>   Nice level. Defaults to 1\n");
fprintf(stderr,"-p <int>   Minimum cycle period (in us). Defaults to <%d>us\n",SLEEPTIME);
          exit(1);
        break; 
        default  : fprintf(stderr,"Unknown option -%c. Try -h for help\n",ptr[0]);
          exit(1);
        break;
      }
      ptr++;
    }
#ifdef TRACE 
fprintf(stderr,"Option <%s>\n",ptr);
#endif   
    i++;
  }
  if(niceval>0&&niceval<40)nice(niceval);

  if(!(useleds||usespeaker)){
fprintf(stderr,"%s: Nothing to do\n",argv[0]);
  }

/* go into background */
#ifndef TRACE
  close(STDIN_FILENO);
  close(STDOUT_FILENO);
  i=fork();
  if(i){
    if(i<0){
      fprintf(stderr,"%s: Could not fork\n",argv[0]);
      exit(2);
    }
    else{
      exit(0);
    }
  } 

  if(!force){
    if(!stat(LOCKFILE,&fs)){
      fprintf(stderr,"%s already seems to be running.\nUse -f to override or remove pidfile %s if stale\n",argv[0],LOCKFILE);
      exit(1);
    }
  } 

  cfp=fopen(LOCKFILE,"w");
  if(cfp!=NULL){
    fprintf(cfp,"%d\n",getpid());
    fclose(cfp);
  }

  close(STDERR_FILENO);
#endif

  cfp=fopen(CONSOLE,"w");
  if(cfp!=NULL){
 
    consolenum=fileno(cfp);

    signal(SIGINT,trapkill);
    signal(SIGTERM,trapkill);
    
    j=0;

    while(1){

      loadavg (&load,&prev,&past);
      sockstat(&sock);
      meminfo (&ram,&used,&swap,&paged);

      if(load>maxload)maxload=load;
      if(load<minload)minload=load;

      if(load<=arrestload){

        if(flatline>PAUSETIME){
          flatline--;
        }

        if(sock>maxsock)maxsock=sock;
        if(sock<minsock)minsock=sock;
   
        if(used>maxused)maxused=used;
        if(used<minused)minused=used;

        if(paged>maxpaged)maxpaged=paged;
        if(paged<minpaged)minpaged=paged;

        if(useleds){
          if(oldsock<sock)ioval=ioctl(consolenum,KDSETLED,LED_CAP|LED_SCR);
          else            ioval=ioctl(consolenum,KDSETLED,LED_CAP);
        }

        if(usespeaker){
          ioval=ioctl(consolenum,KIOCSOUND,(unsigned long)(MAXFREQ+(MINFREQ-MAXFREQ)*(1.0-(load/arrestload))));
          
        }

        if(ioval){
          exit(2);
        }

        f=((double)(used-minused)/(double)(maxused-minused+1)+(double)(paged-minpaged)/(double)(maxpaged-minpaged+1)+(double)(sock-minsock)/(double)(maxsock-minsock+1))/3.0;

        usleep(f*sleeptime);

        if(usespeaker){
          ioctl(consolenum,KIOCSOUND,0);
        }

        if(useleds){
          if(load<=prev)ioctl(consolenum,KDSETLED,0);
          else          ioctl(consolenum,KDSETLED,LED_SCR);
        }

        usleep(((3.1-(3.0*f))*sleeptime));
      }
      else{

        if(usespeaker){
          ioctl(consolenum,KIOCSOUND,MAXFREQ);
        }

        if(useleds){
          ioctl(consolenum,KDSETLED,LED_SCR|LED_CAP|LED_NUM);
        }

        sleep(flatline);
        flatline=flatline++;

      }

      oldsock=sock;
  
    }

    if(usespeaker){
      ioctl(consolenum,KIOCSOUND,0);
    }

    if(useleds){
      ioctl(consolenum,KDSETLED,0);
    }

    fclose(cfp);
  }

  unlink(LOCKFILE);

  return 0;
}


