/* This file is copyrighted(c) and may only be used in terms of the LGPL */

#ifndef _kunf_h_
#define _kunf_h_

/*************************************************************************/
/* Needed for the FILE structure *****************************************/

#include <stdio.h>

/*************************************************************************/
/* Version information ***************************************************/

#define KUNFIG_VERSION "0.4"

/*************************************************************************/
/* Modes of opening the kunf database ************************************/

#define KUNFIG_OPEN_NOENV          0x0001  /* don't ever look at env var */
#define KUNFIG_OPEN_DEFAULTINGENV  0x0002  /* look at env if arg fails   */
#define KUNFIG_OPEN_OVERRIDINGENV  0x0004  /* look at env before arg     */

#define KUNFIG_OPEN_SIMPLEACCESS   0x0010  /* no write, raw nor comment  */
#define KUNFIG_OPEN_COMPLEXACCESS  0x0020  /* write, raw and comment     */

#define KUNFIG_OPEN_PRINTERRORS    0x0100  /* print errors to stderr     */
#define KUNFIG_OPEN_LOGERRORS      0x0200  /* log errors to syslog       */
#define KUNFIG_OPEN_PRINTWARNINGS  0x0400  /* print warnings to stderr   */
#define KUNFIG_OPEN_LOGWARNINGS    0x0800  /* log warnings to syslog     */

#define KUNFIG_OPEN_EXITONERRORS   0x1000  /* do exit() if we get errors */
#define KUNFIG_OPEN_EXITONWARNINGS 0x2000  /* do exit() on warnings      */

#define KUNFIG_OPEN_STANDARD       0x1511  /* sensible options           */

/*************************************************************************/
/* Return error codes: In most cases you should not need these ***********/

#define KUNFIG_ERROR_NONE          0x0000  /* everything peachy          */ 

#define KUNFIG_ERROR_MALLOC        0x0001  /* malloc/realloc failed      */
#define KUNFIG_ERROR_EOF           0x0002  /* read error in fgetc()      */
#define KUNFIG_ERROR_SYNTAX        0x0004  /* syntax error               */
#define KUNFIG_ERROR_ACCESS        0x0008  /* could not open a file      */
#define KUNFIG_ERROR_CALL          0x0010  /* other functions failed     */

/*************************************************************************/
/* Make sure that C++ does not mangle the names **************************/

#ifdef __cplusplus
extern "C" {
#endif

/*************************************************************************/
/* You should not access this structure directly *************************/

struct kunfig_item{
  char *name;
  char *rawname;

  char *value;
  char *rawvalue;

  char *desc;

  unsigned int flags;

  struct kunfig_item *next;
  struct kunfig_item *down;
};

typedef struct kunfig_item KITEM;

/*************************************************************************/
/* The 3 functions below are all you need for a simple applications ******/

extern int    kunfig_open          (char *fname, unsigned int flags);
extern int    kunfig_close         ();

extern char  *kunfig_findvalue     (int number, char *first, ...);

/*************************************************************************/
/* These functions could be required for a more complex applications *****/

extern KITEM *kunfig_findentry(int number, char *first, ...);

extern KITEM *kunfig_next     (KITEM *prev);
extern KITEM *kunfig_down     (KITEM *up);   
extern KITEM *kunfig_start    ();           /* get very first/top item   */

extern int    kunfig_isentry  (KITEM *ptr); /* check this before getvalue*/
extern int    kunfig_isheader (KITEM *ptr); /* check this before down()  */

extern char  *kunfig_getname  (KITEM *ptr); /* extract name from KITEM   */
extern char  *kunfig_getvalue (KITEM *ptr); /* extract value from KITEM  */

/*************************************************************************/
/* The functions below should only be neccessary for a browser ***********/

extern int kunfig_isincludestart(KITEM *ptr); /* an include directive    */
extern int kunfig_isincludeend  (KITEM *ptr); /* EOF in an included file */
extern int kunfig_isredirected  (KITEM *ptr); /* a section in a new file */

extern char *kunfig_getrawname    (KITEM *ptr); /* no % escapes in name  */
extern char *kunfig_getrawvalue   (KITEM *ptr); /* no % escapes in value */
extern char *kunfig_getdescription(KITEM *ptr); /* get description       */

extern char *kunfig_getfilename   (KITEM *ptr); /* filename of inc/redir */

/*************************************************************************/
/* The next set of functions should only be required by an editor ********/

extern int kunfig_setrawvalue (KITEM *ptr, char *rawvalue); /* set value */

extern int kunfig_writeback   ();        /* flush modified items to disk */

/*************************************************************************/
/* These 2 functions can be used to figure out errors ********************/

extern int kunfig_printerror(FILE *stream);
extern int kunfig_waserror  ();       /* true if we encountered an error */

/*************************************************************************/
/* You generally should not look at this variable ************************/

extern unsigned int kunfig_error;

/*************************************************************************/
/* Make sure that C++ does not mangle the names **************************/

#ifdef __cplusplus
        }
#endif

#endif
