/*** This file is copyrighted (c) and may only be used in terms of the LGPL ***/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <syslog.h>

#include <stdarg.h>

#include <sys/file.h>

/******************************************************************************/

#include "kunf.h"
#include "internal_kunf.h"

/******************************************************************************/

unsigned int kunfig_error=KUNFIG_ERROR_NONE;

/******************************************************************************/

KITEM       *_kkunfig_root=NULL;
unsigned int _kkunfig_flags  =0;

/******************************************************************************/

KITEM *kunfig_next(KITEM *prev){
  KITEM *result;

  result=prev->next;

  return result;
}

KITEM *kunfig_down(KITEM *up){
  KITEM *result=NULL;

  if(up->flags&KUNFIGDOWN){
    if(up->flags&KUNFIGUNKNOWN){
      _kkunfig_build(up->value,up);
    }
    result=up->down;
  }

  return result;
}

KITEM *kunfig_start(){
  if(_kkunfig_root==NULL){
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: API usage error: start() called after failed open(). Potential for seg faults");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: API usage error: start() called after failed open(). Potential for seg faults\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_CALL;
    return NULL;
  }
  else{
    return _kkunfig_root->down;
  }
}

int kunfig_isentry(KITEM *ptr){
  if(ptr==NULL){
    return 0;
  }
  else{
    if(ptr->flags&KUNFIGPLAIN){
      return 1;
    }
    else{
      return 0;
    }
  }
}

int kunfig_isheader(KITEM *ptr){
  if(ptr==NULL){
    return 0;
  }
  else{
    if(ptr->flags&KUNFIGDOWN){
      return 1;
    }
    else{
      return 0;
    }
  }
}

int kunfig_isincludestart(KITEM *ptr){
  if(ptr!=NULL&&(ptr->flags&KUNFIGSTARTINC)){
    return 1;
  }
  return 0;
}

int kunfig_isincludeend(KITEM *ptr){
  if(ptr!=NULL&&(ptr->flags&KUNFIGENDINC)){
    return 1;
  }
  return 0;
}

int kunfig_isredirected(KITEM *ptr){
  if(ptr!=NULL&&(ptr->flags&KUNFIGDOWN)&&(ptr->value!=NULL)){
    return 1;
  }
  return 0;
}

char *kunfig_getfilename(KITEM *ptr){
  if(ptr!=NULL&&(ptr->flags&(KUNFIGDOWN|KUNFIGINC))&&ptr->value!=NULL){
    return ptr->value;
  }
  return NULL;
}

char *kunfig_getname(KITEM *ptr){
  char *result=NULL;
  if(ptr!=NULL)result=ptr->name;
  return result;
}

char *kunfig_getvalue(KITEM *ptr){
  char *result=NULL;
  if(ptr!=NULL)result=ptr->value;
  return result;
}

char *kunfig_getrawname(KITEM *ptr){
  char *result=NULL;
  if(ptr!=NULL)result=ptr->rawname;
  return result;
}

char *kunfig_getrawvalue(KITEM *ptr){
  char *result=NULL;
  if(ptr!=NULL)result=ptr->rawvalue;
  return result;
}

char *kunfig_getdescription(KITEM *ptr){
  char *result=NULL;
  if(ptr!=NULL)result=ptr->desc;
  return result;
}

KITEM *kunfig_findentry   (int number, char *first, ...){
  KITEM  *result;
  va_list ap;
  int     count;
  char   *next;

  count=number-1;

  result=kunfig_start();

  while(result!=NULL&&(result->flags&KUNFIGINC||strcmp(first,result->name))){
    result=result->next;
  }
  
  va_start(ap, first);
  while(count>0){
    if(result!=NULL){
      result=kunfig_down(result);
    }

    if(result!=NULL){
      next=va_arg(ap, char *);

      while(result!=NULL&&(result->flags&KUNFIGINC||strcmp(next,result->name))){
        result=result->next;
      }

      count--;
    }
    else{
      count=0;
    }
  }
  va_end(ap);

  return result;
}

char *kunfig_findvalue   (int number, char *first, ...){
  KITEM  *result;
  va_list ap;
  int     count;
  char   *next;

  count=number-1;

  result=kunfig_start();

  while(result!=NULL&&(result->flags&KUNFIGINC||strcmp(first,result->name))){
    result=result->next;
  }
  
  va_start(ap, first);
  while(count>0){
    if(result!=NULL){
      result=kunfig_down(result);
    }

    if(result!=NULL){
      next=va_arg(ap, char *);

      while(result!=NULL&&(result->flags&KUNFIGINC||strcmp(next,result->name))){
        result=result->next;
      }

      count--;
    }
    else{
      count=0;
    }
  }
  va_end(ap);

  if(result!=NULL&&result->flags&KUNFIGPLAIN){
    return result->value;
  }
  else{
    return NULL;
  }

}

int    kunfig_open        (char *fname, unsigned int flags){
  char *name,*eptr;


  if(flags==0){
    _kkunfig_flags=KUNFIG_OPEN_STANDARD;   
  }
  else{
    _kkunfig_flags=flags;   
  }
 
  if(flags&KUNFIG_OPEN_OVERRIDINGENV){
    eptr=getenv(KUNFIGENV);
    if(eptr==NULL){
      name=fname;
    }
    else{
      name=eptr;
    }
  }
  else if(flags&KUNFIG_OPEN_DEFAULTINGENV){
    if(fname==NULL){
      name=getenv(KUNFIGENV);
    }
    else{
      name=fname;
    }
  }
  else{
    name=fname;
  }

  if(_kkunfig_startscan()){
#ifdef TRACE
fprintf(stderr,"open(): Could not boot scanner\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM for scanner");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM for scanner\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
  }

  if(_kkunfig_startescape()){
#ifdef TRACE
fprintf(stderr,"open(): Could not boot escaping system\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not set up escapes");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not set up escapes\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
  }

#ifdef TRACE
fprintf(stderr,"open(): Will start with file <%s>\n",name==NULL?KUNFIGROOT:name);
#endif

  if((_kkunfig_root=_kkunfig_new(NULL,name==NULL?KUNFIGROOT:name,NULL,KUNFIGDOWN|KUNFIGUNKNOWN))==NULL){
#ifdef TRACE
fprintf(stderr,"open(): Could not allocate root\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
  }
  else{
    _kkunfig_build(_kkunfig_root->value,_kkunfig_root);
#ifdef TRACE
    _kkunfig_dump(_kkunfig_root);
#endif
  }

  return 0;
}

int    kunfig_close       (){

  _kkunfig_quitscan();
  _kkunfig_quitescape();
  _kkunfig_free(_kkunfig_root);

  return 0;
}

int    kunfig_printerror  (FILE *stream){
  switch(kunfig_error){
    case KUNFIG_ERROR_MALLOC :
fprintf(stream,"kunf: Could not allocate RAM\n");
    break;
    case KUNFIG_ERROR_EOF :
fprintf(stream,"kunf: Encountered error while reading an input stream\n");
    break;
    case KUNFIG_ERROR_SYNTAX :
fprintf(stream,"kunf: Encountered a syntax error\n");
    break;
    case KUNFIG_ERROR_ACCESS :
fprintf(stream,"kunf: Could not access a file\n");
    break;
    case KUNFIG_ERROR_CALL :
fprintf(stream,"kunf: Could not complete a system or library call\n");
    break;
  }
  return 0;
}

int    kunfig_waserror    (){
  return kunfig_error;
}


int    kunfig_setrawvalue(KITEM *ptr, char *rawvalue){
  char *rawvaluestr;
  char *valuestr;

  if(_kkunfig_flags&KUNFIG_OPEN_COMPLEXACCESS&&rawvalue!=NULL&&(!(ptr->flags&KUNFIGDOWN))){
    rawvaluestr=strdup(rawvalue);
    valuestr=_kkunfig_doescape(rawvalue);

    if(rawvaluestr==NULL||valuestr==NULL){
      if(rawvaluestr!=NULL)free(rawvaluestr);
      if(valuestr!=NULL)   free(valuestr);
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM for modified value");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM for modified value\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
      return kunfig_error;
    }
    else{
      free(ptr->value);
      free(ptr->rawvalue);
      ptr->value   =valuestr;
      ptr->rawvalue=rawvaluestr;
      ptr->flags   =ptr->flags|KUNFIGDIRTY;
    }
  }
  else{
#ifdef TRACE
fprintf(stderr,"setrawvalue(): Usage error. You need to use the OPEN_COMPLEXACCESS flag\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: API usage error: Attempting to modify header or values in OPEN_SIMPLEACCESS mode");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: API usage error: Attempting to modify header or values in OPEN_SIMPLEACCESS mode\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_CALL;
    return kunfig_error;
  }

  return 0;
}

int kunfig_writeback(){
  KITEM **stack;
  KITEM    *tmp;
  int  stackptr;
  int  stackmax;
  int  stackoff;
  int         i;
  FILE      *fp;

  stackmax=KUNFSTACK;
  stackptr=0;

#ifdef TRACE
fprintf(stderr,"writeback(): kunfig_error=<%d>\n",kunfig_error);
#endif

  stack=(KITEM **)malloc(sizeof(KITEM *)*stackmax);

  if(_kkunfig_flags&KUNFIG_OPEN_COMPLEXACCESS){

  if(stack!=NULL){
#ifdef TRACE
fprintf(stderr,"writeback(): Could allocate stack\n");
#endif

    tmp=_kkunfig_root;
    stack[0]=tmp;
    stackptr=1;
    
    do{
      while(tmp!=NULL){
        if(tmp->flags&KUNFIGDIRTY){
          if(!(tmp->flags&(KUNFIGDOWN|KUNFIGINC))){
            i=stackptr;
            do{
              i--;
              if(stack[i]->value!=NULL){
                stack[i]->flags=stack[i]->flags|KUNFIGDIRTY;
#ifdef TRACE
fprintf(stderr,"writeback(): Marked file <%s> as dirty\n",stack[i]->rawvalue);
#endif
                i=0;
              }
            }while(i>0);
          }
        }
        if(tmp->flags&KUNFIGDOWN){
          if(tmp->flags&(KUNFIGFAILED|KUNFIGUNKNOWN)){
            tmp=tmp->next;
          }
          else{
/* push */
#ifdef TRACE
fprintf(stderr,"writeback(): Pushed <%s>\n",tmp->name);
#endif
            stack[stackptr]=tmp;
            stackptr++;
            tmp=tmp->down;
          }
        }
        else{
          if(tmp->flags&KUNFIGSTARTINC){
            stack[stackptr]=tmp;
            stackptr++;
          }
          else if(tmp->flags&KUNFIGENDINC){
            stackptr--;
            if(stack[stackptr]->flags&KUNFIGDIRTY){
/* print include here */
              _kkunfig_writeinclude(stack[stackptr]);
            }
          }
          tmp=tmp->next;
        }
      }

      stackptr--;

      if(stack[stackptr]->flags&KUNFIGDIRTY){
        stack[stackptr]->flags=stack[stackptr]->flags&(~KUNFIGDIRTY);

/* it might be a good idea to do some sort of locking here */                
        fp=fopen(stack[stackptr]->value,"w");
        if(fp!=NULL){
#ifdef TRACE
fprintf(stderr,"writeback(): File <%s> has been modified\n",stack[stackptr]->value);
#endif
          tmp=stack[stackptr];
          tmp=tmp->down;
          stackptr++;
          stackoff=stackptr;

          do{ 
            while(tmp!=NULL){
              if(tmp->flags&KUNFIGDOWN){
fputc('[',fp);
                if(tmp->flags&(KUNFIGFAILED|KUNFIGUNKNOWN)||(tmp->value!=NULL)){
#ifdef TRACE
fprintf(stderr,"writeback(): Skipping <%s>\n",tmp->name);
#endif
                  i=stackoff;
                  while(i<stackptr){
_kkunfig_escapeprint(stack[i]->rawname,fp);
fputc(':',fp);
                    i++;
                  }
_kkunfig_escapeprint(tmp->rawname,fp);
fputc('=',fp);
_kkunfig_escapeprint(tmp->value,fp);

                  tmp=tmp->next;
                }
                else{
#ifdef TRACE
fprintf(stderr,"writeback(): While writing pushed <%s>\n",tmp->name);
#endif

                  tmp->flags=tmp->flags&(~KUNFIGDIRTY);
                  stack[stackptr]=tmp;
                  stackptr++;

                  i=stackoff;
                  while(i<stackptr){
_kkunfig_escapeprint(stack[i]->rawname,fp);
                    i++;
                    if(i<stackptr){
fputc(':',fp);
                    }
                  }

#ifdef TRACE
fprintf(stderr,"writeback(): Persuing <%s> down\n",tmp->name);
#endif
                  tmp=tmp->down;
                }
fprintf(fp,"]\n");
              }
              else{
                if(tmp->flags&KUNFIGSTARTINC){
                  if(!(stack[stackptr-1]->flags&KUNFIGINC)){
fputc('<',fp);
_kkunfig_escapeprint(tmp->value,fp);
fputc('>',fp);
                  }
                  stack[stackptr]=tmp;
                  stackptr++;
                }
                else if(tmp->flags&KUNFIGENDINC){
                  stackptr--;
                }
                else{
                  tmp->flags=tmp->flags&(~KUNFIGDIRTY);
                  if(!(stack[stackptr-1]->flags&KUNFIGINC)){
_kkunfig_escapeprint(tmp->rawname,fp);
fputc('=',fp);
_kkunfig_escapeprint(tmp->rawvalue,fp);
if(tmp->desc!=NULL){
  fprintf(fp," :");
  _kkunfig_escapeprint(tmp->desc,fp);
}
fprintf(fp,"\n");
                  }
                }
                tmp=tmp->next;
              }
            }
#ifdef TRACE
fprintf(stderr,"writeback(): While writing about to pop <%s>\n",stack[stackptr-1]->name);
#endif
            stackptr--;
            tmp=stack[stackptr];
            tmp=tmp->next;
#ifdef TRACE
if(tmp!=NULL)fprintf(stderr,"writeback(): Now working with <%s>\n",tmp->name);
else         fprintf(stderr,"writeback(): Empty.\n");
#endif

          }while(stackptr>=stackoff);
          fclose(fp);
        }
        else{
if(_kkunfig_flags&KUNFIG_OPEN_LOGWARNINGS)syslog(KUNFWLOG,"kunfig: Could not open file <%s> for writing",stack[stackptr]->value);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTWARNINGS)fprintf(stderr,"kunfig: Could not open file <%s> for writing\n",stack[stackptr]->value);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONWARNINGS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_ACCESS;
        }
      }


      tmp=stack[stackptr];
      tmp=tmp->next;
#ifdef TRACE
fprintf(stderr,"writeback(): Popped <%s>\n",stack[stackptr]->name);
#endif

        
    }while(stackptr>0);
    
    free(stack);
  }
  else{
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM for for writing back");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM for for writing back\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
  }

  }
  else{

if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: API usage error: You can not write changes in OPEN_SIMPLEACCESS mode");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: API usage error: You can not write changes in OPEN_SIMPLEACCESS mode\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_CALL;

  }

#ifdef TRACE
  _kkunfig_dump(_kkunfig_root);
#endif

  return kunfig_error;  
}
