/*** This file is copyrighted (c) and may only be used in terms of the LGPL ***/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.h>

#include <unistd.h>
#include <pwd.h>
#include <syslog.h>

#include <sys/utsname.h>
#include <sys/types.h>

#include "kunf.h"
#include "internal_kunf.h"

int _kkunfig_dump(KITEM *top){
  KITEM *i,*j,*k,*l;

  if(top!=NULL){
fprintf(stderr,"<%s=%s:%s> \t(%s=%s) \t0x%x\n",top->name,top->value,top->desc,top->rawname,top->rawvalue,top->flags);
    i=top->down;
    while(i!=NULL){
fprintf(stderr,"  <%s=%s:%s> \t(%s=%s) \t0x%x\n",i->name,i->value,i->desc,i->rawname,i->rawvalue,i->flags);
      j=i->down;
      while(j!=NULL){
fprintf(stderr,"    <%s=%s:%s> \t(%s=%s) \t0x%x\n",j->name,j->value,j->desc,j->rawname,j->rawvalue,j->flags);
        k=j->down;
        while(k!=NULL){
fprintf(stderr,"      <%s=%s:%s> \t(%s=%s) \t0x%x\n",k->name,k->value,k->desc,k->rawname,k->rawvalue,k->flags);
          l=k->down;
          while(l!=NULL){
fprintf(stderr,"        <%s=%s:%s> \t(%s=%s) \t0x%x\n",l->name,l->value,l->desc,l->rawname,l->rawvalue,l->flags);
            l=l->next;
          }
          k=k->next;
        }
        j=j->next;
      }
      i=i->next;
    }
  }

  return 0;
}

int _kkunfig_escapeprint(char *str, FILE *fp){
  int i,j;

  if(str!=NULL){

    for(i=0,j=1;(str[i]!='\0'&&j);i++){
      if(!(isalnum(str[i])||str[i]==':'||str[i]=='_'))j=0;
    }
  
    if(j){
      fprintf(fp,"%s",str);
    }
    else{
      fputc('"',fp);
      for(i=0;str[i]!='\0';i++){
        switch(str[i]){
          case '"' :
          case '\\':
            fputc('\\',fp);
          break;
        }
        fputc(str[i],fp);
      }
      fputc('"',fp);
    }
  }

  return 0;
}


int _kkunfig_writeinclude(KITEM *include){
  int count=0;
  KITEM *tmp;
  FILE *fp;

/* we should do some proper locking hereabouts */

  fp=fopen(include->value,"w");
  if(fp!=NULL){

#ifdef TRACE
fprintf(stderr,"writeinclude(): Will write include file <%s>\n",include->value);
#endif

    tmp=include->next;
    while(tmp!=NULL&&(count>=0)){
      if(count==0){
        if(tmp->flags&KUNFIGSTARTINC){
fputc('<',fp);
_kkunfig_escapeprint(tmp->rawvalue,fp);
fprintf(fp,">\n");
        }
        else if(tmp->flags&KUNFIGPLAIN){
_kkunfig_escapeprint(tmp->rawname,fp);
fputc('=',fp);
_kkunfig_escapeprint(tmp->rawvalue,fp);
if(tmp->desc!=NULL){
  fprintf(fp," :");
  _kkunfig_escapeprint(tmp->desc,fp);
}
fprintf(fp,"\n");
        }
        tmp->flags=tmp->flags&(~KUNFIGDIRTY);
      }
      if(tmp->flags&KUNFIGENDINC){
#ifdef TRACE
fprintf(stderr,"writeinclude(): Encountered end of include <%s> with count=<%d>\n",tmp->value,count);
#endif
        count--;
      }
      else if(tmp->flags&KUNFIGSTARTINC){
        count++;
      }
      tmp=tmp->next;
    }
#ifdef TRACE
fprintf(stderr,"writeinclude(): Wrote include file <%s>\n",include->value);
#endif
    fclose(fp);
  }
  else{
if(_kkunfig_flags&KUNFIG_OPEN_LOGWARNINGS)syslog(KUNFWLOG,"kunfig: Could not open file <%s> for writing",include->value);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTWARNINGS)fprintf(stderr,"kunfig: Could not open file <%s> for writing\n",include->value);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONWARNINGS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_ACCESS;
  }

  return kunfig_error;
}

int _kkunfig_insert(KITEM *child, KITEM *parent){

/* we could have a pedandic mode here to check for duplicates */

  KITEM *tmp;
  if(parent->down==NULL){
    parent->down=child;
  }
  else{
     tmp=parent->down;
     while(tmp!=NULL){
       if(tmp->next==NULL){
         tmp->next=child;
         tmp=NULL;
       }
       else{
         tmp=tmp->next;
       }
     }
  }
  return 0;
}

KITEM *_kkunfig_include(char *fname, KITEM *prev){
  FILE *fp;
  KITEM *result=NULL;
  int state=1;
  char *str, *name, *tmp;
  KITEM *tmpitem, *tmptail;

  static int deapth=KUNFMAXDEAPTH;

#ifdef TRACE
fprintf(stderr,"include(): Starting with file <%s>\n",fname);
#endif

  deapth--;

  if(deapth>0){

  if((fp=fopen(fname,"r"))==NULL){

#ifdef TRACE
fprintf(stderr,"include(): Could not open file <%s>\n",fname);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGWARNINGS)syslog(KUNFWLOG,"kunfig: Could not open file <%s> for reading",fname);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTWARNINGS)fprintf(stderr,"kunfig: Could not open file <%s> for reading\n",fname);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONWARNINGS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_ACCESS;
  result=prev;
  }
  else{
    result=prev;
    name=NULL;
    
    while(state){
      str=_kkunfig_scan(fp);
      if(str[0]=='\0'){
#ifdef TRACE
fprintf(stderr,"include(): Encountered \\0 from scan()\n");
#endif
        state=0;
      }
      else{
        switch(state){
          case 1 :
            switch(str[0]){
              case '<' :
#ifdef TRACE
fprintf(stderr,"include(): Encountered an include directive\n");
#endif
                state=20;
              break;
              default  : 
#ifdef TRACE
fprintf(stderr,"include(): Encountered a plain entry\n");
#endif
                name=strdup(str);
                if(name==NULL){
#ifdef TRACE
fprintf(stderr,"include(): Could not allocate RAM for name\n");
#endif
                  result=NULL;
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
                  state=0;
                }
                else{
                  state=10;
                }
              break;
            }
          break;
          case 10:
            if(str[0]=='='){
              state=11;
            }
            else{
#ifdef TRACE
fprintf(stderr,"include(): Syntax error in state 10 with <%s>\n",str);
#endif
              result=NULL;
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Encountered a syntax error near token <%s>",str);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Encountered a syntax error near token <%s>\n",str);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_SYNTAX;
              state=0;
            }
          break;
          case 11:
            tmpitem=_kkunfig_new(name,str,NULL,KUNFIGPLAIN);
            free(name);
            name=NULL;
            if(tmpitem==NULL){
#ifdef TRACE
fprintf(stderr,"include(): Could not allocate a structure\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
              result=NULL;
              state=0;
            }
            else{
              result->next=tmpitem;
              result=tmpitem;
              state=12;
            }
          break;
          case 12:
            switch(str[0]){
              case '<' :
#ifdef TRACE
fprintf(stderr,"include(): Encountered an include directive\n");
#endif
                state=20;
              break;
              case ':' :
#ifdef TRACE
fprintf(stderr,"include(): Encountered a comment\n");
#endif
                state=13;
              break; 
              default  : 
#ifdef TRACE
fprintf(stderr,"include(): Encountered a plain entry\n");
#endif
                name=strdup(str);
                if(name==NULL){
#ifdef TRACE
fprintf(stderr,"include(): Could not allocate RAM for name\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
                  state=0;
                }
                else{
                  state=10;
                }
              break;
            }
          break;
          case 13:
            tmp=strdup(str);
            if(tmp==NULL){
#ifdef TRACE
fprintf(stderr,"include(): Could not allocate RAM for a comment\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
              result=NULL;
              state=0;
            }
            else{
              tmpitem->desc=tmp;
              state=1;
            }
          break;
          case 20:
            if(_kkunfig_flags&KUNFIG_OPEN_COMPLEXACCESS){
              tmpitem=_kkunfig_new(NULL,str,NULL,KUNFIGSTARTINC);
              tmptail=_kkunfig_new(NULL,str,NULL,KUNFIGENDINC);
              if(tmpitem==NULL||tmptail==NULL){
#ifdef TRACE
fprintf(stderr,"include(): Could not allocate RAM for include file\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
                result=NULL;
                state=0;
              }
              else{
                result->next=tmpitem;
                result=_kkunfig_include(str,tmpitem);
                if(result==NULL){
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not process include file");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not process include file\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_ACCESS;
                  state=0;
                  _kkunfig_free(tmptail);
                }
                else{
                  result->next=tmptail;
                  result=tmptail;
                  state=21;
                }
              }
            }
            else{
              result=_kkunfig_include(str,result);
              if(result==NULL){
                state=0;
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not process include file");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not process include file\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_ACCESS;
              }
              else{
                state=21;
              }
            }
          break;
          case 21:
            if(str[0]=='>'){
              state=1;
            }
            else{
#ifdef TRACE
fprintf(stderr,"build(): Syntax error in state 21 with <%s>\n",str);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Encountered a syntax error near token <%s>",str);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Encountered a syntax error near token <%s>\n",str);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_SYNTAX;
              state=0;
            }
          break;
        }
      }
    }
    fclose(fp);
  }

  }
  else{
#ifdef TRACE
fprintf(stderr,"include(): Nesting of deapth greater than %d\n",KUNFMAXDEAPTH);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Nesting too deep");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Nesting too deep\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_SYNTAX;
  }

  deapth++;

#ifdef TRACE
fprintf(stderr,"include(): Done\n");
#endif

  return result; 
}

KITEM *_kkunfig_build(char *fname, KITEM *top){
  FILE *fp;
  KITEM *result=NULL;
  int state=1;
  char *str, *name, *tmp;
  KITEM *ctop;
  KITEM *ctmp,*tmptail;
  KITEM firstinclude;
  static int deapth=KUNFMAXDEAPTH;

#ifdef TRACE
fprintf(stderr,"build(): Starting\n");
#endif

  deapth--;

  if(deapth>0){

  top->flags=top->flags&(~KUNFIGUNKNOWN);

  if((fp=fopen(fname,"r"))==NULL){

#ifdef TRACE
fprintf(stderr,"build(): Could not open file <%s>\n",fname);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGWARNINGS)syslog(KUNFWLOG,"kunfig: Could not open file <%s> for reading",fname);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTWARNINGS)fprintf(stderr,"kunfig: Could not open file <%s> for reading\n",fname);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONWARNINGS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_ACCESS;
  }
  else{
    ctop=top; 
    name=NULL;
    
    while(state){
      str=_kkunfig_scan(fp);
      if(str[0]=='\0'){
#ifdef TRACE
fprintf(stderr,"build(): Encountered \\0 from scan()\n");
#endif
        state=0;
      }
      else{
        switch(state){
          case 1 :
            switch(str[0]){
              case '<' :
#ifdef TRACE
fprintf(stderr,"build(): Encountered an include directive\n");
#endif
                state=20;
              break;
              case '[' :
#ifdef TRACE
fprintf(stderr,"build(): Encountered a new section\n");
#endif
                ctop=top;
                state=30;
              break; 
              default  : 
#ifdef TRACE
fprintf(stderr,"build(): Encountered a plain entry\n");
#endif
                name=strdup(str);
                if(name==NULL){
#ifdef TRACE
fprintf(stderr,"build(): Could not allocate RAM for name\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
                  state=0;
                }
                else{
                  state=10;
                }
              break;
            }
          break;
          case 10:
            if(str[0]=='='){
              state=11;
            }
            else{
#ifdef TRACE
fprintf(stderr,"build(): Syntax error in state 10 with <%s>\n",str);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Encountered a syntax error near token <%s>",str);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Encountered a syntax error near token <%s>\n",str);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_SYNTAX;
              state=0;
            }
          break;
          case 11:
            ctmp=_kkunfig_new(name,str,NULL,KUNFIGPLAIN);
            free(name);
            name=NULL;
            if(ctmp==NULL){
#ifdef TRACE
fprintf(stderr,"build(): Could not allocate a structure\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
              state=0;
            }
            else{
              _kkunfig_insert(ctmp,ctop);
              state=12;
            }
          break;
          case 12:
            switch(str[0]){
              case '<' :
#ifdef TRACE
fprintf(stderr,"build(): Encountered an include directive\n");
#endif
                state=20;
              break;
              case '[' :
#ifdef TRACE
fprintf(stderr,"build(): Encountered a new section\n");
#endif
                ctop=top;
                state=30;
              break; 
              case ':' :
#ifdef TRACE
fprintf(stderr,"build(): Encountered a comment\n");
#endif
                state=13;
              break; 
              default  : 
#ifdef TRACE
fprintf(stderr,"build(): Encountered a plain entry\n");
#endif
                name=strdup(str);
                if(name==NULL){
#ifdef TRACE
fprintf(stderr,"build(): Could not allocate RAM for name\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
                  state=0;
                }
                else{
                  state=10;
                }
              break;
            }
          break;
          case 13:
            tmp=strdup(str);
            if(tmp==NULL){
#ifdef TRACE
fprintf(stderr,"build(): Could not allocate RAM for a comment\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
              state=0;
            }
            else{
              ctmp->desc=tmp;
              state=1;
            }
          break;


          case 20:
            if(_kkunfig_flags&KUNFIG_OPEN_COMPLEXACCESS){
              ctmp   =_kkunfig_new(NULL,str,NULL,KUNFIGSTARTINC);
              tmptail=_kkunfig_new(NULL,str,NULL,KUNFIGENDINC);
              if(ctmp==NULL||tmptail==NULL){
#ifdef TRACE
fprintf(stderr,"build(): Could not allocate RAM for a file entry\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
                result=NULL;
                state=0;
              }
              else{
                _kkunfig_insert(ctmp,ctop);
                ctmp=_kkunfig_include(str,ctmp);
                if(ctmp==NULL){
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not process include file");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not process include file\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_ACCESS;
                  state=0;
                  result=NULL;
                }
                else{
                  ctmp->next=tmptail;
                  state=21;
                }
              }
            }
            else{
              if(_kkunfig_include(str,&firstinclude)==NULL){
                result=NULL;
                state=0;
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not process include file");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not process include file\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_ACCESS;
              }
              else{
                if(firstinclude.next!=NULL){
                  _kkunfig_insert(firstinclude.next,ctop);
                }
                state=21;
              }
            }
          break;
          case 21:
            if(str[0]=='>'){
              state=1;
            }
            else{
#ifdef TRACE
fprintf(stderr,"build(): Syntax error in state 21 with <%s>\n",str);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Encountered a syntax error near token <%s>",str);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Encountered a syntax error near token <%s>\n",str);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_SYNTAX;
              state=0;
            }
          break;
          case 30:
            if(ctop->down==NULL){
              ctop->down=_kkunfig_new(str,NULL,NULL,KUNFIGDOWN);
              if(ctop->down==NULL){
#ifdef TRACE
fprintf(stderr,"build(): Memory allocation error\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
                state=0;
              }
              else{
#ifdef TRACE
fprintf(stderr,"build(): Creating and descending from <%s> into <%s>\n",ctop->name,ctop->down->name);
#endif
                ctop=ctop->down;
                state=31;
              }
            }
            else{
              ctmp=ctop->down;
              while(ctmp!=NULL){
                if(ctmp->name!=NULL&&(strcmp(ctmp->name,str)==0)){
                  if(ctmp->flags&KUNFIGDOWN){
#ifdef TRACE
fprintf(stderr,"build(): Descending from <%s> into <%s>\n",ctop->name,ctmp->name);
#endif
                    ctop=ctmp;
                    state=31;
                  }
                  else{
                    state=0;
/* flag errors here */
#ifdef TRACE
fprintf(stderr,"build(): Header already in use as a value\n");
#endif
                  }
                  ctmp=NULL;
                }
                else{
                  if(ctmp->next==NULL){
                    ctmp->next=_kkunfig_new(str,NULL,NULL,KUNFIGDOWN);
                    if(ctmp->next==NULL){
#ifdef TRACE
fprintf(stderr,"build(): Memory allocation error\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
                      state=0;
                    }
                    else{
#ifdef TRACE
fprintf(stderr,"build(): Creating and descending from <%s> into <%s>\n",ctop->name,ctmp->name);
#endif
                      ctop=ctmp->next;
                      state=31;
                    }
                    ctmp=NULL;
                  }
                  else{
                    ctmp=ctmp->next;
                    state=31;
                  }
                }
              }
            }
          break;
          case 31:
            switch(str[0]){
              case ']' :
                state=1;
              break;
              case ':' :
                state=30;
              break;
              case '=' :
                state=32;
              break;
            }
          break;
          case 32:
            ctop->flags=ctop->flags|KUNFIGUNKNOWN;
            ctop->value=_kkunfig_doescape(str);
            if(ctop->value==NULL){
#ifdef TRACE
fprintf(stderr,"build(): Could not allocate RAM\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
              state=0;
            }
            else{
              if(_kkunfig_flags&KUNFIG_OPEN_COMPLEXACCESS){
                ctop->rawvalue=strdup(str);
                if(str==NULL){
#ifdef TRACE
fprintf(stderr,"build(): Could not allocate RAM\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
                  state=0;
                }
                else{
                  state=33;
                }
              }
              else{
                state=33;
              }
            }
          break;
          case 33:
            if(str[0]==']'){
              state=34;
            }
            else{
#ifdef TRACE
fprintf(stderr,"build(): Syntax Error\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Encountered a syntax error near token <%s>",str);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Encountered a syntax error near token <%s>\n",str);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_SYNTAX;
              state=0;
            }
          break;
          case 34:
            if(str[0]=='['){
              state=30;
              ctop=top;
            }
            else{
#ifdef TRACE
fprintf(stderr,"build(): Syntax Error\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Encountered a syntax error near token <%s>",str);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Encountered a syntax error near token <%s>\n",str);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_SYNTAX;
              state=0;
            }
          break;
        }
      }
    }
    fclose(fp);
  }

  }
  else{
#ifdef TRACE
fprintf(stderr,"build(): Nesting of deapth greater than %d\n",KUNFMAXDEAPTH);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Nesting too deep");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Nesting too deep\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_SYNTAX;
  }

#ifdef TRACE
fprintf(stderr,"build(): Done\n");
#endif

  return result; 
}

/* eager beaver table construction */

int   _kkunfig_infolen[256];
char *_kkunfig_infostr[256];
char *_kkunfig_infobuf=NULL;

int  _kkunfig_startescape(){

  int i,j;
  char tmp[32];
  char *ptr;

  struct utsname ut;
  struct passwd *pw;

  int uid,gid,euid,egid,pid;

  for(i=0;i<256;i++){
    _kkunfig_infostr[i]=NULL;
    _kkunfig_infolen[i]=0;
  }

  uid=getuid();
  gid=getgid();

  euid=geteuid();
  egid=getegid();

  pid=getpid();
  
  if(uname(&ut)){
#ifdef TRACE
fprintf(stderr,"startescape(): Could not do a uname()\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Call to uname() failed");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Call to uname() failed\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_CALL;
    return 1;
  }

  if((pw=getpwuid(getuid()))==NULL){
#ifdef TRACE
fprintf(stderr,"startescape(): Could not do a getpwuid(%d)\n",uid);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Call to getpwuid(%d) failed",uid);
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Call to getpwuid(%d) failed\n",uid);
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_CALL;
  }
  else{

  if(pw->pw_name)         /*      %l */
    _kkunfig_infolen['l']=(strlen(pw->pw_name)+1);
    
  if(pw->pw_gecos)        /*      %N */
    _kkunfig_infolen['N']=(strlen(pw->pw_gecos)+1);
    
  if(pw->pw_dir)          /*      %d */
    _kkunfig_infolen['d']=(strlen(pw->pw_dir)+1);
    
  if(pw->pw_shell)        /*      %s */
    _kkunfig_infolen['s']=(strlen(pw->pw_shell)+1);

  }

  sprintf(tmp,"%d",uid);  /* UID: %U */
  _kkunfig_infolen['U']=(strlen(tmp)+1);
  
  sprintf(tmp,"%d",gid);  /* GID: %G */
  _kkunfig_infolen['G']=(strlen(tmp)+1);
  
  sprintf(tmp,"%d",euid); /* EUID:%u */
  _kkunfig_infolen['u']=(strlen(tmp)+1);
  
  sprintf(tmp,"%d",egid); /* EGID:%g */
  _kkunfig_infolen['g']=(strlen(tmp)+1);
  
  sprintf(tmp,"%d",pid);  /* PID: %p */
  _kkunfig_infolen['p']=(strlen(tmp)+1);

  if(ut.sysname)          /*      %o */
    _kkunfig_infolen['o']=(strlen(ut.sysname)+1);
        
  if(ut.nodename)         /*      %H */
    _kkunfig_infolen['H']=(strlen(ut.nodename)+1);
        
  if(ut.release)          /*      %R */
    _kkunfig_infolen['R']=(strlen(ut.release)+1);
        
  if(ut.version)          /*      %V */
    _kkunfig_infolen['V']=(strlen(ut.version)+1);
        
  if(ut.machine)          /*      %A */
    _kkunfig_infolen['A']=(strlen(ut.machine)+1);
        
  if(ut.domainname)       /*      %D */
    _kkunfig_infolen['D']=(strlen(ut.domainname)+1);

  _kkunfig_infolen['%']=2; /*     %% */

  _kkunfig_infolen['\0']=1;/*     %\0*/
        
  j=0;
  for(i=0;i<256;i++){
    j+=_kkunfig_infolen[i];
  }

#ifdef TRACE
fprintf(stderr,"startescape(): I will need <%d> bytes\n",j);
#endif

  _kkunfig_infobuf=(char *)malloc(sizeof(char)*j);
  if(_kkunfig_infobuf==NULL){
#ifdef TRACE
fprintf(stderr,"startescape(): Could not allocate <%d> bytes\n",j);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
    return 1;
  }

  ptr=_kkunfig_infobuf;

  sprintf(ptr,"%d",uid);  /* UID: %U */
  _kkunfig_infostr['U']=ptr; ptr=ptr+_kkunfig_infolen['U'];
  
  sprintf(ptr,"%d",gid);  /* GID: %G */
  _kkunfig_infostr['G']=ptr; ptr=ptr+_kkunfig_infolen['G'];
  
  sprintf(ptr,"%d",euid); /* EUID:%u */
  _kkunfig_infostr['u']=ptr; ptr=ptr+_kkunfig_infolen['u'];
  
  sprintf(ptr,"%d",egid); /* EGID:%g */
  _kkunfig_infostr['g']=ptr; ptr=ptr+_kkunfig_infolen['g'];
  
  sprintf(ptr,"%d",pid);  /* PID: %p */
  _kkunfig_infostr['p']=ptr; ptr=ptr+_kkunfig_infolen['p'];

  if(pw!=NULL&&pw->pw_name){/*    %l */
    _kkunfig_infostr['l']=ptr; strcpy(ptr,pw->pw_name); ptr=ptr+_kkunfig_infolen['l'];
  }
    
  if(pw!=NULL&&pw->pw_gecos){/*   %N */
    _kkunfig_infostr['N']=ptr; strcpy(ptr,pw->pw_gecos); ptr=ptr+_kkunfig_infolen['N'];
  }
    
  if(pw!=NULL&&pw->pw_dir){/*     %d */
    _kkunfig_infostr['d']=ptr; strcpy(ptr,pw->pw_dir); ptr=ptr+_kkunfig_infolen['d'];
  }
    
  if(pw!=NULL&&pw->pw_shell){/*   %s */
    _kkunfig_infostr['s']=ptr; strcpy(ptr,pw->pw_shell); ptr=ptr+_kkunfig_infolen['s'];
  }

  if(ut.sysname){         /*      %o */
    _kkunfig_infostr['o']=ptr; strcpy(ptr,ut.sysname); ptr=ptr+_kkunfig_infolen['o'];
  }
        
  if(ut.nodename){        /*      %H */
    _kkunfig_infostr['H']=ptr; strcpy(ptr,ut.nodename); ptr=ptr+_kkunfig_infolen['H'];
  }
        
  if(ut.release){         /*      %R */
    _kkunfig_infostr['R']=ptr; strcpy(ptr,ut.release); ptr=ptr+_kkunfig_infolen['R'];
  }
        
  if(ut.version){         /*      %V */
    _kkunfig_infostr['V']=ptr; strcpy(ptr,ut.version); ptr=ptr+_kkunfig_infolen['V'];
  }
        
  if(ut.machine){         /*      %A */
    _kkunfig_infostr['A']=ptr; strcpy(ptr,ut.machine); ptr=ptr+_kkunfig_infolen['A'];
  }
        
  if(ut.domainname){      /*      %D */
    _kkunfig_infostr['D']=ptr; strcpy(ptr,ut.domainname); ptr=ptr+_kkunfig_infolen['D'];
  }

  strcpy(ptr,"%");        /*      %% */
  _kkunfig_infostr['%']=ptr; ptr=ptr+2;

  strcpy(ptr,"");         /*      %\0*/
  _kkunfig_infostr['\0']=ptr; ptr=ptr+1;

  return 0;  
}

char *_kkunfig_doescape(char *str){
  char *result;
  int      i,j;

#ifdef TRACE
fprintf(stderr,"doescape(): About to process <%s>\n",str);
#endif

  j=1;
  i=0;
  while(str[i]!='\0'){
    if(str[i]=='%'){
      i++;
      j+=_kkunfig_infolen[(int)str[i]];
    }
    else{
      i++;
      j++;
    }
  }

#ifdef TRACE
fprintf(stderr,"doescape(): Need <%d> bytes\n",j);
#endif

  result=(char *)malloc(sizeof(char)*j);
  if(result==NULL){
#ifdef TRACE
fprintf(stderr,"doescape(): Could not allocate <%d> bytes\n",j);
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
  }
  else{
     
    j=0;
    i=0;
    while(str[i]!='\0'){
      if(str[i]=='%'){
#ifdef TRACE
fprintf(stderr,"doescape(): Will add <%s>\n",_kkunfig_infostr[(int)str[i+1]]);
#endif
        if(_kkunfig_infostr[(int)str[i+1]]==NULL){
          result[j]=str[i];
          j++;
        }
        else{
#ifdef TRACE
fprintf(stderr,"doescape(): Will append <%s:%d> at index <%d>\n",_kkunfig_infostr[(int)str[i+1]],_kkunfig_infolen[(int)str[i+1]],j);
#endif
          strcpy(result+j,_kkunfig_infostr[(int)str[i+1]]);
          j+=(_kkunfig_infolen[(int)str[i+1]])-1;
          if(str[i+1]!='\0')i++;
        }
      }
      else{
        result[j]=str[i];
        j++;
      }
      i++;
    }
    result[j]='\0';

  }


  return result;
}

int _kkunfig_quitescape(){
#ifdef TRACE
fprintf(stderr,"quitescape(): Shutting down\n");
#endif
  if(_kkunfig_infobuf!=NULL){
    free(_kkunfig_infobuf);
    _kkunfig_infobuf=NULL;
  }
#ifdef TRACE
fprintf(stderr,"quitescape(): Managed to shut down\n");
#endif
  return 0;
}

KITEM *_kkunfig_new(char *name, char *value, char *desc, unsigned flags){
  KITEM *result;
  int failure=0;

#ifdef TRACE
fprintf(stderr,"new(): Starting\n");
#endif

  result=(KITEM *)malloc(sizeof(KITEM));
  if(result!=NULL){
    result->name=NULL;
    result->rawname=NULL;

    result->value=NULL;
    result->rawvalue=NULL;

    result->desc=NULL;

    result->flags=flags;

    result->next=NULL;
    result->down=NULL;

    if(name!=NULL){
      result->name=_kkunfig_doescape(name);
      if(result->name==NULL){
#ifdef TRACE
fprintf(stderr,"new(): Could not allocate RAM\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
        failure=1;
      }
    }
    if(value!=NULL){
      result->value=_kkunfig_doescape(value);
      if(result->value==NULL){
#ifdef TRACE
fprintf(stderr,"new(): Could not allocate RAM\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
        failure=1;
      }
    }
    if(desc!=NULL&&(_kkunfig_flags&KUNFIG_OPEN_COMPLEXACCESS)){
      result->desc=_kkunfig_doescape(desc);
      if(result->desc==NULL){
#ifdef TRACE
fprintf(stderr,"new(): Could not allocate RAM\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
        failure=1;
      }
    }
    if(_kkunfig_flags&KUNFIG_OPEN_COMPLEXACCESS){
      result->flags=(result->flags|KUNFIGHASRAW);
      if(name!=NULL){
        result->rawname=strdup(name);
        if(result->rawname==NULL){
#ifdef TRACE
fprintf(stderr,"new(): Could not allocate RAM\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
          failure=1;
        }
      }
      if(value!=NULL){
        result->rawvalue=strdup(value);
        if(result->rawvalue==NULL){
#ifdef TRACE
fprintf(stderr,"new(): Could not allocate RAM\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
          failure=1;
        }
      }
    }

    if(failure){
      if(result->name!=NULL)    free(result->name);
      if(result->rawname!=NULL) free(result->rawname);
      if(result->value!=NULL)   free(result->value);
      if(result->rawvalue!=NULL)free(result->rawvalue);
      if(result->desc!=NULL)    free(result->desc);
      free(result);
    }

  }
  else{
#ifdef TRACE
fprintf(stderr,"new(): Could not allocate RAM\n");
#endif
if(_kkunfig_flags&KUNFIG_OPEN_LOGERRORS)syslog(KUNFLOG,"kunfig: Could not allocate RAM");
if(_kkunfig_flags&KUNFIG_OPEN_PRINTERRORS)fprintf(stderr,"kunfig: Could not allocate RAM\n");
if(_kkunfig_flags&KUNFIG_OPEN_EXITONERRORS)exit(KUNFIGEXIT);
kunfig_error=kunfig_error|KUNFIG_ERROR_MALLOC;
  }

#ifdef TRACE
fprintf(stderr,"new(): Allocated <%p> for <%s>\n",result,result->name);
#endif

  return result;
}

int _kkunfig_free(KITEM *ptr){
  KITEM *tmp,*bck;

#ifdef TRACE
fprintf(stderr,"free(): About to deallocate <%p>\n",ptr);
#endif

  if(ptr!=NULL){

    tmp=ptr->down;
    while(tmp!=NULL){
      bck=tmp->next;
      _kkunfig_free(tmp);
      tmp=bck;
    }

    if(ptr->name!=NULL)    free(ptr->name);
    if(ptr->rawname!=NULL) free(ptr->rawname);
    if(ptr->value!=NULL)   free(ptr->value);
    if(ptr->rawvalue!=NULL)free(ptr->rawvalue);
    if(ptr->desc!=NULL)    free(ptr->desc);
    free(ptr);

  }
  return 0;
}

#ifdef BOOTSTRAP
int main(int argc,char **argv){

  _kkunfig_startescape();

  if(argc>1)printf("doescape(): <%s>\n",_kkunfig_doescape(argv[1]));

  _kkunfig_quitescape();
  return 0;
}
#endif
