/*** This file is copyrighted(c) any may only be used in terms of the GPL ***/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <syslog.h>

#include <kunf.h>

#define  BUFMAX 2048 

/****************************************************************************/

int printformata(KITEM *cptr, char *sptr, int sindex);
int printformatb(KITEM *cptr, char *sptr, int sindex);
int printformatc(KITEM *cptr, char *sptr, int sindex);
int printformatd(KITEM *cptr, char *sptr, int sindex);
int escapeprint (char *str);
int mycpy       (char *dst, char *src, int offset);

/****************************************************************************/

int main(int argc, char **argv){

  char   sptr[BUFMAX];
  char   *fname=NULL;
  char   *ptr;

  int     i,j;

  int     mode=4;   

  i=1;

  while(i<argc){
    ptr=argv[i];
    if(ptr[0]=='-'){
      j=1;
      while(ptr[j]!='\0'){
        switch(ptr[j]){
          case 'H' :
          case 'h' :
printf("%s: Print all configuration data to stdout.\n[tool for the KUNFIG library v%s]\n\n",argv[0],KUNFIG_VERSION);
printf("Usage: %s [-hilp] [kunfigfile]\n\n",argv[0]);
printf(" -h            : this help\n");
printf(" -i            : indented formatting (according to deapth)\n");
printf(" -l            : line formatting (each line contains section, name and value)\n");
printf(" -p            : plain formatting (section and name joined with underscore and\n       only values are escaped)\n\n");
          exit(1);
          break;
          case 'p' :
            mode=2;
          break;
          case 'i' :
            mode=3;
          break;
          case 'l' :
            mode=1; 
          break;
          case '-' :
          break;
printf("%s: Unrecognized option -%c\n",argv[0],ptr[j]);
        }
        j++;
      }
    }
    else{
      if(fname!=NULL){
fprintf(stderr,"%s: Ignoring parameter <%s>\n",argv[0],fname);
      }
      fname=ptr;
    }
    i++;
  }

  if(kunfig_open(fname,KUNFIG_OPEN_DEFAULTINGENV|KUNFIG_OPEN_SIMPLEACCESS|KUNFIG_OPEN_PRINTERRORS|KUNFIG_OPEN_EXITONERRORS|KUNFIG_OPEN_PRINTWARNINGS)){
fprintf(stderr,"%s: Could not initialize\n",argv[0]);
    exit(2);
  }

  sptr[0]='\0'; 

  switch(mode){
    case 1 :
      printformata(kunfig_start(),sptr,0);
    break;
    case 2 :
      printformatb(kunfig_start(),sptr,0);
    break;
    case 3 :
      printformatc(kunfig_start(),sptr,0);
    break;
    case 4 :
      printformatd(kunfig_start(),sptr,0);
    break;
  }

  kunfig_close();

  if(kunfig_waserror()){
    return 2;
  }

  return 0;
}

int printformata(KITEM *cptr, char *sptr, int sindex){
  int tmp;

  while(cptr!=NULL){
    if(kunfig_isentry(cptr)){
      if(sindex>0){
        putchar('[');
        escapeprint(sptr);
        printf("] ");
      }
      escapeprint(kunfig_getname(cptr));
      putchar('=');
      escapeprint(kunfig_getvalue(cptr));
      putchar('\n');
    }
    else if(kunfig_isheader(cptr)){
#ifdef TRACE
fprintf(stderr,"printrow(): Descending using header <%s>\n",kunfig_getname(cptr));
#endif
      if(sindex>0){
        sptr[sindex]=':';
        tmp=mycpy(sptr+sindex+1,kunfig_getname(cptr),sindex);
        printformata(kunfig_down(cptr),sptr,tmp+1);
      }
      else{
        tmp=mycpy(sptr,kunfig_getname(cptr),0);
        printformata(kunfig_down(cptr),sptr,tmp);
      }
      sptr[sindex]='\0';
    }
    cptr=kunfig_next(cptr);
  }

  return 0;
}

int printformatb(KITEM *cptr, char *sptr, int sindex){
  int tmp;

  while(cptr!=NULL){
    if(kunfig_isentry(cptr)){
      if(sindex>0){
        printf("%s_%s=",sptr,kunfig_getname(cptr));
      }
      else{
        printf("%s=",kunfig_getname(cptr));
      }
      escapeprint(kunfig_getvalue(cptr));
      putchar('\n');
    }
    else if(kunfig_isheader(cptr)){
#ifdef TRACE
fprintf(stderr,"printrow(): Descending using header <%s>\n",kunfig_getname(cptr));
#endif
      if(sindex>0){
        sptr[sindex]='_';
        tmp=mycpy(sptr+sindex+1,kunfig_getname(cptr),sindex);
        printformatb(kunfig_down(cptr),sptr,tmp+1);
      }
      else{
        tmp=mycpy(sptr,kunfig_getname(cptr),0);
        printformatb(kunfig_down(cptr),sptr,tmp);
      }
      sptr[sindex]='\0';
    }
    cptr=kunfig_next(cptr);
  }

  return 0;
}

int printformatc(KITEM *cptr, char *sptr, int sindex){
  static d=0;
  int i;
  int tmp;

  d+=2;

  while(cptr!=NULL){
    if(kunfig_isentry(cptr)){
      for(i=0;i<d;i++)putchar(' ');
      escapeprint(kunfig_getname(cptr));
      putchar('=');
      escapeprint(kunfig_getvalue(cptr));
      putchar('\n');
      cptr=kunfig_next(cptr);
    }
    else if(kunfig_isheader(cptr)){
#ifdef TRACE
fprintf(stderr,"printrow(): Descending using header <%s>\n",kunfig_getname(cptr));
#endif
      if(sindex>0){
        sptr[sindex]=':';
        tmp=mycpy(sptr+sindex+1,kunfig_getname(cptr),sindex);
        for(i=0;i<d;i++)putchar(' '); 
        printf("[%s]\n",sptr);
        printformatc(kunfig_down(cptr),sptr,tmp+1);
      }
      else{
        tmp=mycpy(sptr,kunfig_getname(cptr),0);
        printf("  [%s]\n",sptr);
        printformatc(kunfig_down(cptr),sptr,tmp);
      }
      sptr[sindex]='\0';
      cptr=kunfig_next(cptr);
      if(sindex>0&&cptr!=NULL){
        for(i=0;i<d-2;i++)putchar(' ');
        printf("[%s]\n",sptr);
      }
    }
    else{
      cptr=kunfig_next(cptr);
    }
  }

  d-=2;

  return 0;
}

int printformatd(KITEM *cptr, char *sptr, int sindex){
  int tmp;

  while(cptr!=NULL){
    if(kunfig_isentry(cptr)){
      
      printf("  ");
      escapeprint(kunfig_getname(cptr));
      putchar('=');
      escapeprint(kunfig_getvalue(cptr));
      putchar('\n');
      
      cptr=kunfig_next(cptr);
    }
    else if(kunfig_isheader(cptr)){
#ifdef TRACE
fprintf(stderr,"printrow(): Descending using header <%s>\n",kunfig_getname(cptr));
#endif
      if(sindex>0){
        sptr[sindex]=':';
        tmp=mycpy(sptr+sindex+1,kunfig_getname(cptr),sindex);
        if(kunfig_down(cptr)!=NULL&&(!kunfig_isheader(kunfig_down(cptr)))){
          printf("[%s]\n",sptr);
        }
        printformatd(kunfig_down(cptr),sptr,tmp+1);
      }
      else{
        tmp=mycpy(sptr,kunfig_getname(cptr),0);
        if(kunfig_down(cptr)!=NULL&&(!kunfig_isheader(kunfig_down(cptr)))){
          printf("[%s]\n",sptr);
        }
        printformatd(kunfig_down(cptr),sptr,tmp);
      }
      sptr[sindex]='\0';
      cptr=kunfig_next(cptr);
      if(sindex>0&&cptr!=NULL){
        printf("[%s]\n",sptr);
      }
    }
    else{
      cptr=kunfig_next(cptr);
    }
  }

  return 0;
}

int escapeprint(char *str){
  int i,j;

  if(str!=NULL){

    for(i=0,j=1;(str[i]!='\0'&&j);i++){
      if(!(isalnum(str[i])||str[i]==':'||str[i]=='_'))j=0;
    }
  
    if(j){
      printf("%s",str);
    }
    else{
      putchar('"');
      for(i=0;str[i]!='\0';i++){
        switch(str[i]){
          case '"' :
          case '\\':
            putchar('\\');
          break;
        }
        putchar(str[i]);
      }
      putchar('"');
    }
  }

  return 0;
}

int mycpy       (char *dst, char *src, int offset){
  
  while(offset<BUFMAX&&(*src!='\0')){
    offset++;
    *dst=(*src);
    dst++;
    src++;
  }

  if(*src=='\0'){
    *dst='\0';
  }
  else{
fprintf(stderr,"kunfcat: Buffer overrrun. This should not happen. Your headers are far, far, far too long and/or deep. I'll report that to the syslog.\n");
syslog(LOG_ERR|LOG_PID|LOG_USER,"kunfcat: Buffer overrun. Deliberate ?\n");
    exit(2);
  }
  
  return offset;
}

