/*** This file is copyrighted(c) any may only be used in terms of the GPL ***/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include <ncurses.h>

#include <kunf.h>

#define MODE_SHOWVALUE        0x02
#define MODE_SHOWRAWVALUE     0x03
#define MODE_SHOWDESCRIPTION  0x04

#define SBCOL 1
#define CVCOL 2
#define FRCOL 3

#define PRIMARY    "administration"
#define SECONDARY  "kunf"
#define TERTIARY   "edit"

#define FGSBCOL "statusbar_foreground"
#define BGSBCOL "statusbar_background"
#define FGCVCOL "canvas_foreground"
#define BGCVCOL "canvas_background"
#define FGFRCOL "frame_foreground"
#define BGFRCOL "frame_background"

int colok=0;
int haschanged=0;

struct level{
  KITEM **ptr;
  int    cptr;
  int    iptr;
  int    optr;
  int    mlen;
};

typedef struct level MLEV;

int   docolpair  (int pair, char *fg, char *bg);
int   setupcolors();
int   doalevel   (MLEV *m, int xoff);
MLEV *makealevel (KITEM *first);
int   zapalevel  (MLEV *m);
int   drawalevel (MLEV *m, int xoff, int mode);
int   matchcol   (char *str);
char *editw      (int y, int x, char *str);

int matchcol(char *str){
  if(     !strcasecmp("black",str)){
    return COLOR_BLACK;
  }
  else if(!strcasecmp("green",str)){
    return COLOR_GREEN;
  }
  else if(!strcasecmp("yellow",str)){
    return COLOR_YELLOW;
  }
  else if(!strcasecmp("blue",str)){
    return COLOR_BLUE;
  }
  else if(!strcasecmp("cyan",str)){
    return COLOR_CYAN;
  }
  else if(!strcasecmp("magenta",str)){
    return COLOR_MAGENTA;
  }
  else if(!strcasecmp("white",str)){
    return COLOR_WHITE;
  }
  else{
    return COLOR_RED;
  }
}

int docolpair(int pair, char *fg, char *bg){
  init_pair(pair, (fg==NULL)?COLOR_WHITE:matchcol(fg), (bg==NULL)?COLOR_BLUE:matchcol(bg));
  return 0;
}

int setupcolors(){
  char *fg, *bg;

#ifdef TRACE
fprintf(stderr,"setupcolors(): Setting up colors\n");
#endif
  
/*  if(has_colors()==TRUE&&can_change_color()==TRUE&&COLOR_PAIRS>5){*/
  if(has_colors()==TRUE&&COLOR_PAIRS>5){
    fg=kunfig_findvalue(4,PRIMARY,SECONDARY,TERTIARY,FGSBCOL);
    bg=kunfig_findvalue(4,PRIMARY,SECONDARY,TERTIARY,BGSBCOL);

    docolpair(SBCOL,fg,bg);

    fg=kunfig_findvalue(4,PRIMARY,SECONDARY,TERTIARY,FGCVCOL);
    bg=kunfig_findvalue(4,PRIMARY,SECONDARY,TERTIARY,BGCVCOL);

    docolpair(CVCOL,fg,bg);

    fg=kunfig_findvalue(4,PRIMARY,SECONDARY,TERTIARY,FGFRCOL);
    bg=kunfig_findvalue(4,PRIMARY,SECONDARY,TERTIARY,BGFRCOL);

    docolpair(FRCOL,fg,bg);

    colok=1;
#ifdef TRACE
fprintf(stderr,"setupcolors(): Done setting up colors\n");
#endif
  }
#ifdef TRACE
sleep(2);
#endif
  return 0;
}

MLEV *makealevel(KITEM *first){
  KITEM *tmp;
  MLEV  *result;
  int    i,j=0;
  char  *tstr;

  result=(MLEV *)malloc(sizeof(MLEV));
  if(result!=NULL){

    result->iptr=0;
    result->optr=0;
    result->mlen=0;
    
    i=0;
    tmp=first;
    while(tmp!=NULL){
      i++;
      tstr=kunfig_getname(tmp);
      if(tstr!=NULL){
        j=strlen(tstr);
        if(j>result->mlen)result->mlen=j;
      }
      tmp=kunfig_next(tmp);
    }

    result->cptr=i;

    result->ptr=(KITEM **)malloc(sizeof(KITEM *)*i);
    if(result->ptr!=NULL){
      tmp=first;
      i=0;
      while(tmp!=NULL){
        result->ptr[i]=tmp;
        tmp=kunfig_next(tmp);
        i++;
      }
    }
    else{
      free(result);
      result=NULL;
fprintf(stderr,"makealevel(): Could not allocate RAM\n");
    }
      
  }
  else{
fprintf(stderr,"makealevel(): Could not allocate RAM\n");
  }

  return result;
};

int doalevel(MLEV *m, int xoff){
  int  inp,run;
  MLEV      *n;
  KITEM     *p;

  int      z,x;
  char      *s;

  run=1;

/* do status line here */
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
if(m->cptr==1)mvprintw(LINES-1,0,"  1 entry"); 
else          mvprintw(LINES-1,0,"  %d entries",m->cptr); 
if(colok)attrset(COLOR_PAIR(CVCOL));

  do{
    drawalevel(m,xoff,MODE_SHOWRAWVALUE|MODE_SHOWDESCRIPTION);
    
    inp=getch();

    switch(inp){
      case 'w'        :
      case 'W'        :
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  writing changes..."); 
        refresh();
        if(kunfig_writeback()){
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  trouble writing changes..."); 
          beep();
        }
        else{
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  wrote changes..."); 
        }
if(colok)attrset(COLOR_PAIR(CVCOL));
        haschanged=0;
      break;
      case '4'        :
      case KEY_LEFT   :
      case KEY_EIC    :
      case KEY_CANCEL :
      case KEY_CLOSE  :
      case 27         :
/*      case 'q'        :*/
/*      case 'Q'        :*/
        run=0;
      break;
/*      case KEY_NPAGE  : m->optr=m->optr+(LINES/2); */
      case '2'        :
      case KEY_DOWN   :
        m->optr=m->optr+1;
        if(m->optr>=m->cptr){
          m->optr=0;
          m->iptr=0;
        }
        else{
          if((m->iptr+(LINES/2))<m->optr){
            m->iptr=m->iptr+1;
          }
        }
      break;
/*      case KEY_PPAGE  : m->optr=m->optr-(LINES/2); */
      case '8'        :
      case KEY_UP     :
        m->optr=m->optr-1;
        if(m->optr<0){
          m->optr=(m->cptr>0)?(m->cptr-1):0;
          if(m->cptr>LINES-2){
            m->iptr=(m->cptr-(LINES/2));
          }
        }
        else{
          if((m->iptr+(LINES/2))>m->optr){
            m->iptr=m->iptr-1;
            if(m->iptr<0)m->iptr=0;
          }
        }
      break;
      case '6'       :
      case KEY_ENTER :
      case KEY_RIGHT :
      case '\n'      :
      case '\r'      :
        if(kunfig_isheader(m->ptr[m->optr])){
          p=kunfig_down(m->ptr[m->optr]);
          if(p!=NULL){
            n=makealevel(p);
            if(n!=NULL){
              s=kunfig_getname(m->ptr[m->optr]);
              if(s==NULL){
                z=xoff+((COLS/2)-xoff)/3;
              }
              else{
                z=strlen(s)+2+xoff;
              }
              if(z<2||z>((COLS/5)*3))z=xoff;
              doalevel(n,z);
              zapalevel(n);
if(colok)attron(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
if(m->cptr==1)mvprintw(LINES-1,0,"  1 entry"); 
else          mvprintw(LINES-1,0,"  %d entries",m->cptr); 
if(colok)attroff(COLOR_PAIR(SBCOL));
            }
          }
          else{
if(colok)attron(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
            if(kunfig_isredirected(m->ptr[m->optr])){
              s=kunfig_getfilename(m->ptr[m->optr]);
              if(s==NULL)mvprintw(LINES-1,0,"  Could not aquire a name for redirected file");
              else       mvprintw(LINES-1,0,"  File <%s> is either empty or inacessible",s); 
            }
            else{
              s=kunfig_getname(m->ptr[m->optr]);
              mvprintw(LINES-1,0,"  Section <%s> is empty",s); 
            }
if(colok)attroff(COLOR_PAIR(SBCOL));
            refresh();
          }
        }
        else if(kunfig_isentry(m->ptr[m->optr])){
/* do status line here */
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  Edit value and type <ESC> when completed");
if(colok)attrset(COLOR_PAIR(CVCOL));

          s=kunfig_getname(m->ptr[m->optr]);
          if(s!=NULL){
            x=strlen(s)+xoff+3;
            if(x>COLS-5)x=xoff;
            s=editw((m->optr)-(m->iptr),x,kunfig_getrawvalue(m->ptr[m->optr]));
            if(s!=NULL&&s[0]!='\0'){
              if(strcmp(s,kunfig_getrawvalue(m->ptr[m->optr]))){
                if(kunfig_setrawvalue(m->ptr[m->optr],s)){
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  Could not muster resources to change value"); 
if(colok)attrset(COLOR_PAIR(CVCOL));
                }
                else{
                  haschanged=1;
                  m->optr=m->optr+1;
                  if(m->optr>=m->cptr){
                    m->optr=0;
                    m->iptr=0;
                  }
                  else{
                    if((m->iptr+(LINES/2))<m->optr){
                      m->iptr=m->iptr+1;
                    }
                  }
                }
              }
              free(s);
            }
            else{
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  Malformed data entry. Retaining old value"); 
if(colok)attrset(COLOR_PAIR(CVCOL));
            }
          }
          else{
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  Insufficient resources to edit this entry"); 
if(colok)attrset(COLOR_PAIR(CVCOL));
          }
        }
      break;
      case 12 :
        setupcolors();
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,(colok)?' '|COLOR_PAIR(SBCOL):' ');
if(m->cptr==1)mvprintw(LINES-1,0,"  1 entry"); 
else          mvprintw(LINES-1,0,"  %d entries",m->cptr); 
if(colok)attrset(COLOR_PAIR(CVCOL));
        touchwin(stdscr);
        clearok(stdscr,TRUE);
      break;
      case KEY_F0  :
      case KEY_F(1):
if(colok)attron(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,(colok)?' '|COLOR_PAIR(SBCOL):' ');
mvprintw(LINES-1,0,"w: write  ^L: redraw  <RETURN>: edit field  <ESC>: exit edit"); 
if(colok)attroff(COLOR_PAIR(SBCOL));
      break;
    }
  
  }while(run);

  return 0;
}

#define EDITWBUF 128
#define PCOUNT   8

char *editw(int y, int x,char *str){
  char *result=NULL;
  int   run=1;
  int   inp,i;
  char  ebuf[EDITWBUF];
  int   index=0;
  int   offset=0;
  int   max=0;
  int   p=PCOUNT;

  if(str!=NULL){
    max=strlen(str);
    if(max>=EDITWBUF){
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  Old value is too long - truncating..."); 
if(colok)attrset(COLOR_PAIR(CVCOL));
      strncpy(ebuf,str,EDITWBUF-1);
      ebuf[EDITWBUF-1]='\0';
    }
    else{
      strcpy(ebuf,str);
    }
  }
  else{
    ebuf[0]='\0';
    max=0;
  }
  if(colok)attron(A_BOLD);
  for(i=0;(i+x)<COLS&&(str[i]!='\0');i++)mvaddch(y,i+x,str[i]);
  for(;(i+x)<COLS;i++)mvaddch(y,i+x,' ');
  move(y,x);
  if(colok)attroff(A_BOLD);
  
  do{
    inp=getch();
    switch(inp){
      case KEY_ENTER  :
      case '\r'       :
      case '\n'       :
        ebuf[0]='\0';
        max=0;
        for(i=x;i<COLS;i++)mvaddch(y,i,' ');
        move(y,x);
        index=0;
        offset=0;
        p=PCOUNT;
      break;
      case KEY_CANCEL :
      case KEY_CLOSE  :
      case KEY_DOWN   :
      case 27         :
        run=0;
      break;
      case KEY_UP     :
        if(str!=NULL){
          max=strlen(str);
          if(max>=EDITWBUF){
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  Old value is too long - truncating..."); 
if(colok)attrset(COLOR_PAIR(CVCOL));
            strncpy(ebuf,str,EDITWBUF-1);
            ebuf[EDITWBUF-1]='\0';
          }
          else{
            strcpy(ebuf,str);
          }
          if(colok)attron(A_BOLD);
          for(i=0;(i+x)<COLS&&(str[i]!='\0');i++)mvaddch(y,i+x,str[i]);
          for(;(i+x)<COLS;i++)mvaddch(y,i+x,' ');
          if(colok)attroff(A_BOLD);
        }
        move(y,x);
        index=0;
        offset=0;
        p=PCOUNT;
      break;
      case KEY_RIGHT  :
        if(index<max){
          index++;
          if(index-offset+x<COLS){
            move(y,index-offset+x);
          }
          else{
            offset++;
            if(colok)attron(A_BOLD);
            for(i=0;(i+x)<COLS&&(ebuf[i+offset]!='\0');i++)mvaddch(y,i+x,ebuf[i+offset]);
            for(;(i+x)<COLS;i++)mvaddch(y,i+x,' ');
            if(colok)attroff(A_BOLD);
          }
        }
      break;
      case KEY_LEFT   :
        if(index>0){
          index--;
          if(index<offset){
            offset--;
            if(colok)attron(A_BOLD);
            for(i=0;(i+x)<COLS&&(ebuf[i+offset]!='\0');i++)mvaddch(y,i+x,ebuf[i+offset]);
            for(;(i+x)<COLS;i++)mvaddch(y,i+x,' ');
            if(colok)attroff(A_BOLD);
          }
          move(y,index-offset+x);
          p=PCOUNT;
        }
        else{
          p--;
          run=p;
        }
      break;
      case KEY_DC     :
        if(colok)attron(A_BOLD);
        if(max>index){
          for(i=index;i<max;i++)ebuf[i]=ebuf[i+1];
          max--;
          for(i=0;(i+x)<COLS&&(ebuf[i+offset]!='\0');i++)mvaddch(y,i+x,ebuf[i+offset]);
          for(;(i+x)<COLS;i++)mvaddch(y,i+x,' ');
          move(y,index-offset+x);
        }
        if(colok)attroff(A_BOLD);
      break;
      case 8             :
      case 127           :
      case KEY_BACKSPACE :
        if(colok)attron(A_BOLD);
        if(index>0){
          index--;
          if(index<offset){
            offset--;
          }
          for(i=index;i<max;i++)ebuf[i]=ebuf[i+1];
          max--;
          for(i=0;(i+x)<COLS&&(ebuf[i+offset]!='\0');i++)mvaddch(y,i+x,ebuf[i+offset]);
          for(;(i+x)<COLS;i++)mvaddch(y,i+x,' ');
          move(y,index-offset+x);
        }
        p=PCOUNT;
        if(colok)attroff(A_BOLD);
      break;
      default         :
        if(max<(EDITWBUF-1)){
          if(colok)attron(A_BOLD);
          for(i=max;i>=index;i--)ebuf[i+1]=ebuf[i];
          ebuf[index]=inp;

          if(x+index>=COLS){
            offset++;
          }
          
          max++;
          index++;

          for(i=0;(i+x)<COLS&&(ebuf[i+offset]!='\0');i++)mvaddch(y,i+x,ebuf[i+offset]);
          for(;(i+x)<COLS;i++)mvaddch(y,i+x,' ');

          move(y,index-offset+x);
          if(colok)attroff(A_BOLD);
        }
        else{
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  Value is too long..."); 
if(colok)attrset(COLOR_PAIR(CVCOL));
        }
      break;
    }
  }while(run);

  result=strdup(ebuf);

  return result;
}

int drawalevel(MLEV *m, int xoff, int mode){
  int i,j,k,l,x;
  char *s;

  i=0;
  j=m->iptr;

  while(i<(LINES-1)){
    if(j<m->cptr){
      s=kunfig_getname(m->ptr[j]);
      if(j==m->optr){
        attron(A_REVERSE);
      }
      if(s==NULL){
        k=xoff;
        mvaddch(i,k++,(colok)?ACS_LTEE|COLOR_PAIR(FRCOL):ACS_LTEE);
        if(kunfig_isincludestart(m->ptr[j])){
          mvaddch(i,k++,(colok)?ACS_RTEE|COLOR_PAIR(FRCOL):ACS_PLUS);
          s=kunfig_getfilename(m->ptr[j]);
          if(s!=NULL){
            for(;k<COLS&&(s[k-(xoff+2)]!='\0');k++){
              mvaddch(i,k,(colok)?s[k-(xoff+2)]|COLOR_PAIR(FRCOL):s[k-(xoff+2)]);
            }
          }
          mvaddch(i,k++,(colok)?ACS_LTEE|COLOR_PAIR(FRCOL):ACS_LTEE);
        }
        for(;k<COLS;k++){
          if(colok)mvaddch(i,k,ACS_HLINE|COLOR_PAIR(FRCOL));
          else     mvaddch(i,k,ACS_HLINE);
        }
      }
      else{
        if(colok)mvaddch(i,xoff,ACS_VLINE|COLOR_PAIR(FRCOL));
        else     mvaddch(i,xoff,ACS_VLINE);
        if(kunfig_isheader(m->ptr[j])){
          if(colok)mvaddch(i,xoff+1,ACS_RARROW|COLOR_PAIR(FRCOL));
          else     mvaddch(i,xoff+1,ACS_RARROW);
        }
        else{
          if(colok)mvaddch(i,xoff+1,' '|COLOR_PAIR(CVCOL));
          else     mvaddch(i,xoff+1,' ');
        }
#ifdef TRACE
if((m->ptr[j])->flags&0x40){
if(colok)mvaddch(i,xoff+1,'*'|COLOR_PAIR(CVCOL));
else     mvaddch(i,xoff+1,'*');
}
#endif
        for(k=xoff+2;(k<COLS&&s[k-(xoff+2)]!='\0');k++){
          if(colok)mvaddch(i,k,s[k-(xoff+2)]|COLOR_PAIR(CVCOL));
          else     mvaddch(i,k,s[k-(xoff+2)]);
        }
        if(mode&(MODE_SHOWVALUE|MODE_SHOWRAWVALUE)&&kunfig_isentry(m->ptr[j])){
          if     (mode&MODE_SHOWVALUE)   s=kunfig_getvalue(m->ptr[j]);
          else if(mode&MODE_SHOWRAWVALUE)s=kunfig_getrawvalue(m->ptr[j]);
          if(s!=NULL&&k<COLS-2){
            if(colok)mvaddch(i,k++,'='|COLOR_PAIR(FRCOL));
            else     mvaddch(i,k++,'=');
            for(l=0;(k<COLS&&s[l]!='\0');k++,l++){
              if(colok)mvaddch(i,k,s[l]|COLOR_PAIR(CVCOL));
              else     mvaddch(i,k,s[l]);
            }
          }
        }
        if(mode&MODE_SHOWDESCRIPTION){
          s=kunfig_getdescription(m->ptr[j]);
          if(s!=NULL&&k<COLS-2){
            if(colok)mvaddch(i,k++,':'|COLOR_PAIR(FRCOL));
            else     mvaddch(i,k++,':');
            for(l=0;(k<COLS&&s[l]!='\0');k++,l++){
              if(colok)mvaddch(i,k,s[l]|COLOR_PAIR(CVCOL));
              else     mvaddch(i,k,s[l]);
            }
          }
        }
        for(;k<COLS;k++){
          if(colok)mvaddch(i,k,' '|COLOR_PAIR(CVCOL));
          else     mvaddch(i,k,' ');
        }
      }
      if(j==m->optr){
        attroff(A_REVERSE);
      }
    }
    else{
      if(colok){
        mvaddch(i,xoff,ACS_VLINE|COLOR_PAIR(FRCOL));
        for(x=xoff+1;x<COLS;x++)mvaddch(i,x,' '|COLOR_PAIR(CVCOL));
      }
      else{
        mvaddch(i,xoff,ACS_VLINE);
        for(x=xoff+1;x<COLS;x++)mvaddch(i,x,' ');
      }
    }
    i++;
    j++;
  }

  if(m->iptr>0){
    if(m->iptr==m->optr)mvaddch(0,xoff,(colok)?ACS_UARROW|A_REVERSE|COLOR_PAIR(FRCOL):ACS_UARROW|A_REVERSE);
    else                mvaddch(0,xoff,(colok)?ACS_UARROW|COLOR_PAIR(FRCOL):ACS_UARROW);
  }
  if((m->cptr-(m->iptr))>LINES-1){
    if(m->iptr+(LINES-2)==m->optr)mvaddch(LINES-2,xoff,(colok)?ACS_DARROW|A_REVERSE|COLOR_PAIR(FRCOL):ACS_DARROW|A_REVERSE);
    else                          mvaddch(LINES-2,xoff,(colok)?ACS_DARROW|COLOR_PAIR(FRCOL):ACS_DARROW);
  }
  refresh();

  return 0; 
}

int zapalevel(MLEV *m){
  if(m!=NULL){
    if(m->ptr!=NULL){
      free(m->ptr);
      m->ptr=NULL;
    }
    free(m);
  }
  return 0;
}

int main(int argc, char **argv){
  int i,j,x; 
  char *ptr;

  char *fname=NULL;

  MLEV *top;

#ifdef TRACE
fprintf(stderr,"main(): Starting\n");
#endif

  i=1;
  while(i<argc){
    ptr=argv[i];

    if(ptr[0]=='-'){
      j=1;
      while(ptr[j]!='\0'){
        switch(ptr[j]){
          case '-' :
          break;  
          case 'H' :
          case 'h' :
printf("%s: Edit configurations interactively.\n[tool for the KUNFIG library v%s]\n\n",argv[0],KUNFIG_VERSION);
printf("Usage: %s [-h] [-f kunfigfile]\n\n",argv[0]);
printf(" -f kunfigfile : alternate kunfig file\n");
printf(" -h            : this help (F1 for online hints)\n\n");
            exit(1);
          break;
          case 'f' :
            if(ptr[j+1]=='\0'){
              if((i+1)<argc){
                i++;
                j=0;
                ptr=argv[i];
                fname=ptr;
              }
              else{
fprintf(stderr,"%s: option -f needs a further argument\n",argv[0]);
                exit(2);
              }
            }
            else{
              j++;
              fname=ptr+j;
#ifdef TRACE
fprintf(stderr,"main(): file is <%s>\n",fname);
#endif
            }
            while(ptr[j]!='\0')j++;
            j--;
          break;
          default  :
fprintf(stderr,"%s: Unknown option -%c\n",argv[0],ptr[j]);
            exit(2);
          break;
        }
        j++;
      }
    }
    i++;
  }

  kunfig_open(fname,KUNFIG_OPEN_NOENV|KUNFIG_OPEN_COMPLEXACCESS|KUNFIG_OPEN_PRINTERRORS|KUNFIG_OPEN_EXITONERRORS);

  initscr(); 
  start_color();
  cbreak(); 
  noecho();
  keypad(stdscr,TRUE);
  typeahead(-1);

  setupcolors();

  if(colok){
    attron(COLOR_PAIR(SBCOL));
    for(i=0;i<LINES;i++)for(x=0;x<COLS;x++)mvaddch(i,x,' ');
  }
  mvprintw(LINES-1,0,"  %s: Utility for the KUNF library v%s",argv[0],KUNFIG_VERSION); 
  if(colok)attroff(COLOR_PAIR(SBCOL));


  top=makealevel(kunfig_start());
  if(top!=NULL){
    doalevel(top,0);
    while(haschanged){
if(colok)attrset(COLOR_PAIR(SBCOL));
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  (w)rite changes or e(x)it ?"); 
      switch(getch()){
        case 'w' :
        case 'W' :
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  writing changes..."); 
          if(kunfig_writeback()){
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  trouble writing changes..."); 
            refresh();
            beep();
            sleep(1);
          }
          else{
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  wrote changes..."); 
            haschanged=0;
          }
        break;
        case 'x' :
        case 'X' :
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  changes not written..."); 
          haschanged=0;
        break;
        default  :
for(x=0;x<COLS;x++)mvaddch(LINES-1,x,' ');
mvprintw(LINES-1,0,"  Not a valid response... Try again"); 
        break;
      }
if(colok)attrset(COLOR_PAIR(CVCOL));
      refresh();
    }
    zapalevel(top);
  }

  endwin();

  kunfig_close(); 
  
#ifdef TRACE
fprintf(stderr,"main(): Stopping\n");
#endif

  printf("\n\n");

  return 0;
}

