/*** This file is copyrighted(c) any may only be used in terms of the GPL ***/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <kunf.h>

/****************************************************************************/

int  putintoenv (char *prefix, char *name, char *value);
int  grabentry  (char *name, int index, KITEM *first);
int  escapeprint(char *str);

/****************************************************************************/

int  outputformat=1;

int main(int argc, char **argv){
  int   i,j; 
  char *ptr;

  char *fname=NULL;

#ifdef TRACE
fprintf(stderr,"main(): Starting\n");
#endif

  i=1;
  while(i<argc){
    ptr=argv[i];

    if(ptr[0]=='-'){
      j=1;
      while(ptr[j]!='\0'){
        switch(ptr[j]){
          case '-' :
          break;  
          case 't' :
          case 'c' :
            outputformat=3;
          break;
          case 'T' :
          case 'C' :
            outputformat=4;
          break;
          case 'B' :
          case 'S' :
          case 'K' :
          case 'Z' :
            outputformat=2;
          break; 
          case 'p' : 
            outputformat=5;
          break;
          case 'P' : 
            outputformat=6;
          break;
          case 'H' :
          case 'h' :
printf("%s: Place configuration data into the environment (use `eval %s`).\n[tool for the KUNFIG library v%s]\n\n",argv[0],argv[0],KUNFIG_VERSION);
printf("Usage: %s [-bBcCP] [-f kunfigfile] [entry ...]\n\n",argv[0]);
printf(" -b            : use SH style syntax to declare variables (default)\n");
printf(" -B            : use SH style syntax to place variables into environment\n");
printf(" -c            : use CSH style syntax to declare variables\n");
printf(" -C            : use CSH style syntax to place variables into environment\n");
printf(" -P            : use PERL style syntax\n");
printf(" -f kunfigfile : alternate kunfig file\n");
printf(" -h            : this help\n\n");
            exit(1);
          break;
          case 'f' :
            if(ptr[j+1]=='\0'){
              if((i+1)<argc){
                i++;
                j=0;
                ptr=argv[i];
                fname=ptr;
              }
              else{
fprintf(stderr,"%s: option -f needs a further argument\n",argv[0]);
                exit(2);
              }
            }
            else{
              j++;
              fname=ptr+j;
#ifdef TRACE
fprintf(stderr,"main(): file is <%s>\n",fname);
#endif
            }
            while(ptr[j]!='\0')j++;
            j--;
          break;
          default  :
fprintf(stderr,"%s: Unknown option -%c\n",argv[0],ptr[j]);
            exit(2);
          break;
        }
        j++;
      }
    }
    i++;
  }

#ifdef TRACE
fprintf(stderr,"main(): Calling open with <%s>\n",fname);
#endif

  kunfig_open(fname,KUNFIG_OPEN_NOENV|KUNFIG_OPEN_SIMPLEACCESS|KUNFIG_OPEN_PRINTERRORS|KUNFIG_OPEN_PRINTWARNINGS|KUNFIG_OPEN_EXITONERRORS);

  i=1;
  while(i<argc){
    ptr=argv[i];

/* make sure we skip the argument to -f */
    if(ptr[0]=='-'){
      j=1;
      while(ptr[j]!='\0'){
        if(ptr[j]=='f'){
          j++;
          if(ptr[j]=='\0'){
            i++;
            j=0;
            ptr=argv[i];
          }
          while(ptr[j]!='\0')j++;
        }
        else{
          j++;
        }
      }
    }

/* Here we get the information */
    else{

#ifdef TRACE
fprintf(stderr,"main(): Examining <%s>\n",ptr);
#endif

      grabentry(ptr,0,kunfig_start());

    }
    i++;
  }

  kunfig_close(); 
  
#ifdef TRACE
fprintf(stderr,"main(): Stopping\n");
#endif

  return 0;
}

int grabentry(char *name, int index, KITEM *first){
  int i,j;

#ifdef TRACE
fprintf(stderr,"grabentry(): Was called with <%s:%d>\n",name,index);
#endif

  j=0;
  i=index;
  while(name[i]!='\0'){
    if(name[i]==':'){
      name[i]='\0';
      j=1;
    }
    else{
      i++;
    }
  }

#ifdef TRACE
fprintf(stderr,"grabentry(): Will match against <%s:%d>\n",name+index,i-index);
#endif

  while(first!=NULL&&strcmp(name+index,kunfig_getname(first))){
    first=kunfig_next(first);
  }
  

  if(first!=NULL){
    if(j){
      name[i]='_';
      grabentry(name,i+1,kunfig_down(first));
    }
    else{
      if(kunfig_isentry(first)){
        putintoenv(NULL,name,kunfig_getvalue(first));
      }
      else if(kunfig_isheader(first)){
        first=kunfig_down(first);
        while(first!=NULL){
          if(kunfig_isentry(first))putintoenv(name,kunfig_getname(first),kunfig_getvalue(first));
          first=kunfig_next(first);
        }
      }
    }
  }

  return 0;
}

int putintoenv(char *prefix, char *name, char *value){
  if(outputformat==3)                      printf("set ");
  if(outputformat==4)                      printf("setenv ");
  if(outputformat==6)                      printf("$");
  if(prefix!=NULL&&outputformat!=5)        printf("%s_",prefix);
  if(outputformat!=5)                      printf("%s",name);
  if(outputformat==1||outputformat==2||outputformat==6)putchar('=');
  else if(outputformat==3||outputformat==4)putchar(' ');
                                           escapeprint(value);
  if(outputformat==2){
                                           printf(";\nexport ");
    if(prefix!=NULL)                       printf("%s_",prefix);
                                           printf("%s",name);
  }
  if(outputformat!=5)                      putchar(';');
                                           putchar('\n');
  return 0;
}

int escapeprint(char *str){
  int i,j;

  if(str!=NULL){

    for(i=0,j=1;(str[i]!='\0'&&j);i++){
      if(!(isalnum(str[i])||str[i]=='_'))j=0;
    }
  
    if(j){
      printf("%s",str);
    }
    else{
      putchar('"');
      for(i=0;str[i]!='\0';i++){
        switch(str[i]){
          case '"' :
          case '\\':
            putchar('\\');
          break;
        }
        putchar(str[i]);
      }
      putchar('"');
    }
  }

  return 0;
}

