/*** This file is copyrighted(c) any may only be used in terms of the GPL ***/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <kunf.h>

#define INC_NEST 12
#define REC_NEST 12

KITEM **stac;
int     stacptr;

int   verbose=0;

KITEM **allockitem(int i){
  KITEM **result;

  result=(KITEM **)malloc(sizeof(KITEM *)*i);
  if(result==NULL){
fprintf(stderr,"allockitem(): Could not allocate RAM\n");
    exit(2);
  }

  return result;
}

KITEM **reallocitem(KITEM **old, int i){
  KITEM **result;

  result=(KITEM **)realloc(old,sizeof(KITEM *)*i);
  if(result==NULL){
fprintf(stderr,"reallockitem(): Could not allocate RAM\n");
    exit(2);
  }

  return result;
}

int recurse(KITEM *ptr){
  KITEM *tmp;
  char  *str;
  int      i;
  int  inc=0;
  static d=REC_NEST;

  int errors=0;

  d--;

  if(d==0){
if(verbose>0)printf("Warning         : Redirection is very deep.\n");
if(verbose>2)printf("  That could make things quite confusing.\n");
    errors++;
  }

  while(ptr!=NULL){
if(verbose>4)if((str=kunfig_getname(ptr))!=NULL)printf("Processing item <%s>...\n",str);

    tmp=kunfig_down(ptr);
    
    if(tmp!=NULL){
      if(kunfig_isredirected(ptr)){
        str=kunfig_getfilename(ptr);
        if(str!=NULL){
if(verbose>1)printf("Processing redirection file <%s>...\n",str);
          if(str[0]!='/'){
if(verbose>0)printf("Warning         : Redirected file <%s> has no absolute path.\n",str);
if(verbose>2)printf("  That could make the file inaccessible and even introduce security problems.\n");
            errors++;
          }
          errors+=recurse(tmp);
if(verbose>1)printf("Completed processing redirection file <%s>.\n",str);
        }
        else{
if(verbose>0)printf("Serious Warning : Could not aquire a filename for a redirected file.\n");
if(verbose>2)printf("  Did you specify a null string ? You are lucky to get as far as this message.\n");
          errors++;
        }
      }
      else{
        errors+=recurse(tmp);
      }
    }
    else if((str=kunfig_getfilename(ptr))!=NULL){
      if(kunfig_isincludestart(ptr)){
        inc++;
        if(inc==INC_NEST){
if(verbose>0)printf("Warning         : Inclusion nesting very deep.\n");
if(verbose>2)printf("  That could make things rather confusing.\n");
          errors++;
        }
if(verbose>1)printf("Processing included file <%s>...\n",str);
        if(str[0]!='/'){
if(verbose>0)printf("Warning         : Included file <%s> has no absolute path.\n",str);
if(verbose>2)printf("  That could make the file inaccessible and even introduce security problems.\n");
          errors++;
        }
      }
      else{
        inc--;
if(verbose>1)printf("Completed processing included file <%s>.\n",str);
      }
    }
    else{
      if((str=kunfig_getname(ptr))!=NULL){
        if(!isalpha(str[0])){
if(verbose>0)printf("Warning         : Entry <%s> does not start with a letter.\n",str);
if(verbose>2)printf("  That might make putting it into the environment difficult.\n");
          errors++;
        }
        i=0;
        while(str[i]!='\0'){
          if(!(isalnum(str[i])||str[i]=='_')){
if(verbose>0)printf("Warning         : Entry <%s> contains nonalphanumeric characters.\n",str);
if(verbose>2)printf("  That might make putting it into the environment difficult.\n");
            errors++;
            break;
          }
          i++;
        }
        if(str[i]=='\0'){
          if(i<3){
if(verbose>0)printf("Picky Warning   : Entry <%s> seems to be somewhat short.\n",str);
if(verbose>2)printf("  That might things a bit cryptic for the user.\n");
            errors++;
          }
          else if(i>32){
if(verbose>0)printf("Picky Warning   : Entry <%s> seems to be somewhat long.\n",str);
if(verbose>2)printf("  That might things a tedious for the user.\n");
            errors++;
          }
        }
      }
      else{
if(verbose>0)printf("Serious Warning : Could not aquire a label.\n");
if(verbose>2)printf("  Did you specify a null string ? You are lucky to get as far as this message.\n");
        errors++;
      }
    }
    
    ptr=kunfig_next(ptr);
  }

  d++;

  return errors;
}

/* Return codes :                              */
/* 0 : no warnings                             */
/* 1 : return from -h option                   */
/* 2 : errors inside this code (usage, RAM...) */
/* 3 : there were warnings                     */
/* 5 : there were errors (reported by library) */

int main(int argc, char **argv){
  int   i,j; 
  char *ptr;
  int   warnings;
  char *fname=NULL;

  verbose=1;


#ifdef TRACE
fprintf(stderr,"main(): Starting\n");
#endif

  i=1;
  while(i<argc){
    ptr=argv[i];

    if(ptr[0]=='-'){
      j=1;
      while(ptr[j]!='\0'){
        switch(ptr[j]){
          case '-' :
          break;  
          case 'H' :
          case 'h' :
printf("%s: Check the configuration database.\n[tool for the KUNFIG library v%s]\n\n",argv[0],KUNFIG_VERSION);
printf("Usage: %s [-hqvvv...] [-f kunfigfile]\n\n",argv[0]);
printf(" -f kunfigfile : alternate kunfig file\n");
printf(" -h            : this help\n");
printf(" -q            : suppress almost all output\n");
printf(" -v            : verbose. More -v's increase verbosity\n\n");
            exit(1);
          break;
          case 'f' :
            if(ptr[j+1]=='\0'){
              if((i+1)<argc){
                i++;
                j=0;
                ptr=argv[i];
                fname=ptr;
              }
              else{
fprintf(stderr,"%s: option -f needs a further argument\n",argv[0]);
                exit(2);
              }
            }
            else{
              j++;
              fname=ptr+j;
#ifdef TRACE
fprintf(stderr,"main(): file is <%s>\n",fname);
#endif
            }
            while(ptr[j]!='\0')j++;
            j--;
          break;
          case 'q' :
            verbose--;
          break;
          case 'V' :
            verbose++;
          case 'v' :
            verbose++;
          break;
          default  :
fprintf(stderr,"%s: Unknown option -%c\n",argv[0],ptr[j]);
            exit(2);
          break;
        }
        j++;
      }
    }
    i++;
  }

#ifdef TRACE
fprintf(stderr,"main(): Calling open with <%s>\n",fname);
#endif

  kunfig_open(fname,KUNFIG_OPEN_NOENV|KUNFIG_OPEN_COMPLEXACCESS|KUNFIG_OPEN_PRINTERRORS|KUNFIG_OPEN_PRINTWARNINGS|KUNFIG_OPEN_EXITONERRORS);

  warnings=recurse(kunfig_start());

  kunfig_close(); 

  if(verbose>0){
    switch(warnings){
      case 0 :
printf("Completed check : There were no dubious constructs\n");
      break;
      case 1 :
printf("Completed check : There was a single warning\n");
      break;
      default:
printf("Completed check : There were %d warnings\n",warnings);
      break;
    }
  }
  
#ifdef TRACE
fprintf(stderr,"main(): Stopping\n");
#endif

  if(warnings>0)warnings=3;

  return warnings;
}

